################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: DQ2JobSplitter.py,v 1.41 2009-07-27 13:03:24 mslater Exp $
###############################################################################
# Athena DQ2JobSplitter

import math, socket, operator, copy, os, StringIO
from functools import reduce

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.GPIDev.Lib.Job import Job
from Ganga.GPIDev.Schema import *

from Ganga.Utility.logging import getLogger

from GangaAtlas.Lib.ATLASDataset import DQ2Dataset
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import *
from Ganga.Utility.Config import getConfig, makeConfig, ConfigError

from Ganga.GPIDev.Credentials_old import GridProxy
gridProxy = GridProxy()

logger = getLogger()

def appendFile(file_path, archive_path):
    file_name = os.path.split(file_path)[1]
    test_area = os.environ['TestArea']
    
    #uncompress
    cmd = "gzip -d %s" % archive_path
    out = commands.getoutput(cmd)
    archive = archive_path.replace(".tar.gz", ".tar")
    
    #create tar ball from file list
    cmd = "cd %s ; cat %s|xargs tar cf event_pick.tar" % (test_area, file_name) 
    #print " Tar cmd %s " %cmd
    out = commands.getoutput(cmd)

    #append
    cmd = "cd %s ; tar Af %s event_pick.tar" % (test_area, archive) 
    #print "Append cmd %s " %cmd
    out = commands.getoutput(cmd)

    #compress
    cmd = "gzip %s " %(archive)
    out = commands.getoutput(cmd)
    
    #remove file
    files = ''
    for line in open(file_path, 'r').readlines():
        files += "%s " % line.strip('\n')

    cmd = " cd %s ; rm event_pick.tar %s; rm %s " %(test_area, file_name, files)
    #print "Delete cmd %s " %cmd
    out = commands.getoutput(cmd)

def dq2_siteinfo(dataset, allowed_sites, locations, udays, faxSites, skipReplicaLookup):

    if faxSites:
        result = dq2_list_locations_siteindex(datasets=dataset, days=udays, replicaList=True, allowed_sites=allowed_sites+locations, fax_sites=faxSites, skipReplicaLookup=skipReplicaLookup)
    else:
        result = dq2_list_locations_siteindex(datasets=dataset, days=udays, replicaList=True, allowed_sites= allowed_sites, fax_sites=faxSites, skipReplicaLookup=skipReplicaLookup)
        
    siteinfo = {}
    for guid, sites in result.iteritems():
        newsites = [ site for site in sites if site in allowed_sites ]
        # Remove inconsistencies
        if faxSites:
            newsites = [ site for site in newsites if site in locations+faxSites ]
        else:
            newsites = [ site for site in newsites if site in locations ]

        for site in newsites:
            if site.find('TAPE')>=0:
                newsites.remove(site)
    
        if not newsites: continue
        newsites.sort()
        sitekey = ':'.join(newsites)
        if sitekey in siteinfo:
            siteinfo[sitekey].append(guid)
        else:
            siteinfo[sitekey] = [ guid ]

    return siteinfo



def lfc_siteinfo(result,allowed_sites):
   
    siteinfo = {}
    for guid, sites in result.iteritems():
        newsites = [ site for site in sites if site in allowed_sites ]
        if not newsites: continue
        newsites.sort()
        sitekey = ':'.join(newsites)
        if sitekey in siteinfo:
            siteinfo[sitekey].append(guid)
        else:
            siteinfo[sitekey] = [ guid ]

    return siteinfo
 
class DQ2JobSplitter(ISplitter):
    '''Dataset driven job splitting'''

    _name = 'DQ2JobSplitter'
    _schema = Schema(Version(1,1), {
        'numfiles'          : SimpleItem(defvalue=0,doc='Number of files per subjob'),
        'numsubjobs'        : SimpleItem(defvalue=0,sequence=0, doc="Number of subjobs"),
        'use_lfc'           : SimpleItem(defvalue = False, doc = 'Use LFC catalog instead of default site catalog/tracker service'),
        'update_siteindex'  : SimpleItem(defvalue = True, doc = 'Update siteindex during job submission to get the latest file location distribution.'),
        'use_blacklist'     : SimpleItem(defvalue = True, doc = 'Use black list of sites create by GangaRobot functional tests.'),
        'filesize'          : SimpleItem(defvalue=0, doc = 'Maximum filesize sum per subjob im MB.'),
        'numevtsperjob'     : SimpleItem(defvalue=0, doc='Number of events per subjob'),
        'numevtsperfile'    : SimpleItem(defvalue=0,doc='Maximum number of events in a file of input dataset'),
        'missing_files'     : SimpleItem(defvalue=[],typelist=['str'],sequence=1,protected=1,doc='List of names that could not be assigned to a subjob'),
        'use_fax'           : SimpleItem(defvalue = False, doc = 'Allow submission to a site although data is not there for FAX usage'),
    })

    _GUIPrefs = [ { 'attribute' : 'numfiles',         'widget' : 'Int' },
                  { 'attribute' : 'numsubjobs',       'widget' : 'Int' },
                  { 'attribute' : 'use_lfc',          'widget' : 'Bool' },
                  { 'attribute' : 'update_siteindex', 'widget' : 'Bool' },
                  { 'attribute' : 'use_blacklist',    'widget' : 'Bool' },
                  { 'attribute' : 'filesize',         'widget' : 'Int' },
                  { 'attribute' : 'numevtsperjob',       'widget' : 'Int' },
                  { 'attribute' : 'numevtsperfile',       'widget' : 'Int' }
                  ]


    def split(self,job):

        logger.debug('DQ2JobSplitter called')

        faxSites = []

        logger.debug('job.inputdata.dataset %s job.backend.site %s ' % (job.inputdata.dataset, job.backend.site))

        if not job.inputdata:
            if (job.application.options.find("%SKIPEVENTS") != 0) or (job.application.options.find("%RNDM") != 0):

                # no inputdata but splitting options in command line
                logger.info ("Splitting on command line options rather than input data...")
            
                # check splitting options
                if self.numsubjobs <= 0:
                    raise ApplicationConfigurationError(None,'DQ2JobSplitter must have numsubjobs specified if not using inputdata')

                subjobs = []
                skipevent = 0
                rndm_seed = 0

                # update random seed
                match_rndm = re.search('%RNDM(:*)(\d*)( |$|\'|\"|;)',job.application.options)
                if match_rndm:
                    rndm_seed = int(match_rndm.group(2))

                # check for skip events
                match_skip = re.search('%SKIPEVENTS(:*)(\d*)( |$|\'|\"|;)',job.application.options)
                
                for i in range(0, self.numsubjobs):

                    j = Job()
                    j.name = job.name
                    j.application   = job.application
                    
                    # update the command line given split options
                    if match_rndm:
                        j.application.options = re.sub(match_rndm.group(0),'%s%s' % (rndm_seed, match_rndm.group(3)),j.application.options)

                    if match_skip:
                        if match_skip.group(2) == '':
                            j.application.options = re.sub(match_skip.group(0),'%s%s' % (skipevent,match_skip.group(3)),j.application.options)
                        else:
                            j.application.options = re.sub(match_skip.group(0),'%s%s' % (skipevent + int(match.group(2)), match_skip.group(3)),j.application.options)
                                                                                                                                                        
                    j.application.run_event   = []
                    j.outputdata    = job.outputdata
                    j.backend       = job.backend
                    j.inputsandbox  = job.inputsandbox
                    j.outputsandbox = job.outputsandbox

                    subjobs.append(j)

                    # update event counters
                    if job.splitter.numevtsperjob > 0:
                        skipevent += job.splitter.numevtsperjob
                        
                    rndm_seed += 1
                    
                return subjobs
    
            else:
                raise ApplicationConfigurationError(None,'DQ2JobSplitter specifed but no input dataset and no splitter options (%SKIPEVENTS and/or %RNDM) in job.application.options')            

        if job.inputdata._name != 'DQ2Dataset'  and job.inputdata._name != 'AMIDataset' and job.inputdata._name != 'EventPicking':
            raise ApplicationConfigurationError(None,'DQ2 Job Splitter requires a DQ2Dataset or AMIDataset or EventPicking as input')

        if not job.backend._name in [ 'LCG', 'CREAM', 'Panda', 'NG' ] and not ( job.backend._name in ['SGE'] and config['ENABLE_SGE_DQ2JOBSPLITTER'] ):
            raise ApplicationConfigurationError(None,'DQ2JobSplitter requires an LCG, CREAM, Panda or NG backend')
        
        if (self.numevtsperjob <= 0 and self.numfiles <=0 and self.numsubjobs <=0 and self.filesize <=0):
            raise ApplicationConfigurationError(None,"Specify one of the parameters of DQ2JobSplitter for job splitting: numsubjobs, numfiles, numevtsperjob")
 
        if (self.numevtsperjob > 0 and job.inputdata._name != 'AMIDataset'):
            raise ApplicationConfigurationError(None,"Event based splitting is supported only for AMIDataset as input dataset type")
        # split options are mutually exclusive
        if ( (self.numfiles > 0 or self.numsubjobs > 0) and self.numevtsperjob > 0):
            raise ApplicationConfigurationError(None,"Split by files (or subjobs) and events can not be defined simultaneously")

        if (job.application.max_events > 0 and self.numevtsperjob > 0):
            raise ApplicationConfigurationError(None,"Split by maximum events and split by events can not be defined simultaneously")
        
        # check correct usage of inputdata.tagdataset and create a mapping from TAG DS to parent DS
        tag_dataset_map = {}
        if job.inputdata and job.inputdata._name == 'DQ2Dataset' and len(job.inputdata.tagdataset) != 0:
            if len(job.inputdata.dataset) != 0:

                if len(job.inputdata.tagdataset) != 1 and len(job.inputdata.tagdataset) != len(job.inputdata.dataset):
                    raise ApplicationConfigurationError(None,"There must be a 1->1 mapping of TAG datasets to parent datasets or only a single TAG dataset")

                index = 0

                for inds in job.inputdata.dataset:

                    # expand container
                    indslist = resolve_container([inds]);

                    for inds2 in indslist:
                        if len(job.inputdata.tagdataset) > 0:
                            tag_dataset_map[inds2] = resolve_container( [job.inputdata.tagdataset[index]] )
                        else:
                            tag_dataset_map[inds2] = resolve_container( [job.inputdata.tagdataset[0]] )

                        # attempt to match on tid names
                        if inds2.find('tid') != -1:
                            tidnum = inds2[ inds2.find('tid'): ]

                            for tagds in tag_dataset_map[inds2]:
                                if tagds.find(tidnum) != -1:
                                    tag_dataset_map[inds2] = [tagds]
                                    break
                        
                    index += 1

            else:
                
                # Use ELSSI to match TAGs if required - nicked from PsubUtils to give more flexibility.

                # set up X509_USER_PROXY if needed
                if 'X509_USER_PROXY' in os.environ.keys():
                    old_proxy = os.environ['X509_USER_PROXY']
                else:
                    old_proxy = ''

                os.environ['X509_USER_PROXY'] = gridProxy.location()

                from pandatools import countGuidsClient
                streamRef = 'StreamAOD_ref'
                if 'collRefName' in job.application.atlas_run_config['input']:
                    streamRef = job.application.atlas_run_config['input']['collRefName']
                elif job.inputdata.tag_coll_ref in ['AOD', 'ESD', 'RAW']:
                    streamRef = "Stream%s_ref" % job.inputdata.tag_coll_ref

                newTagDSList = []

                for tagDS in job.inputdata.tagdataset:
                    logger.warning("Using ELSSI DB to match TAG dataset %s to source datasets..." % tagDS)
                    tagIF = countGuidsClient.countGuidsClient()
                    tagIF.debug = False
                    dsNameForLookUp = re.sub('_tid\d+(_\d+)*$','', tagDS)

                    if dsNameForLookUp != tagDS:
                        logger.warning("Cannot search for individual tid datasets. Defaulting to parent container %s" % dsNameForLookUp)

                    if dsNameForLookUp.endswith("/"):
                        dsNameForLookUp = dsNameForLookUp[:-1]

                    tagResults = tagIF.countGuids(dsNameForLookUp,"", streamRef + ",StreamTAG_ref")
                    
                    if not tagResults:
                        tagResults = tagIF.countGuids(dsNameForLookUp,"", streamRef + ",StreamTAG")
                        if not tagResults:
                            raise ApplicationConfigurationError(None,"Could not find references to TAG dataset %s in ELSSI DB. Try matching from dq2 or using TagPrepare." % dsNameForLookUp)
                                                        
                    # NOTE: The folowing should use the TAG info returned by countGuids but ELSSI DB
                    # is messed up for pre 2011 data. This should be fixed!
                    
                    # find the parent datasets
                    parentGUIDs = []
                    for g in tagResults[1]:
                        if not g[1][0] in parentGUIDs:
                            parentGUIDs.append(g[1][0])

                    from pandatools import Client
                    parentDSList = Client.listDatasetsByGUIDs(parentGUIDs,'')
                    parentDSNameList = []
                    for g in parentDSList[0].keys():
                        if not parentDSList[0][g][0] in parentDSNameList:
                            parentDSNameList.append( parentDSList[0][g][0])

                    # attempt to match on tid names with TAG and parent
                    single_tid = ''
                    if tagDS.find('tid') != -1:
                        tidnum = tagDS[ tagDS.find('tid'): ]

                        for ds in parentDSNameList:
                            if ds.find(tidnum) != -1:
                                single_tid = ds

                    # Fill the inputdata fields
                    for g in parentDSList[0].keys():
                        if single_tid != '' and parentDSList[0][g][0] != single_tid:
                            continue

                        if not g in job.inputdata.guids:
                            job.inputdata.guids.append(g)
                            
                        if not parentDSList[0][g][0] in job.inputdata.dataset:
                            job.inputdata.dataset.append(parentDSList[0][g][0])
                            if single_tid != '':
                                tag_dataset_map[ parentDSList[0][g][0] ] = [tagDS]
                            else:
                                if dsNameForLookUp == tagDS:
                                    tag_dataset_map[ parentDSList[0][g][0] ] = resolve_container( [dsNameForLookUp ] )                                    
                                else:
                                    tag_dataset_map[ parentDSList[0][g][0] ] = resolve_container( [dsNameForLookUp + '/'] )



                        if not parentDSList[0][g][1] in job.inputdata.names:
                            job.inputdata.names.append(parentDSList[0][g][1])

                # reset the gird proxy
                if old_proxy != '':
                    os.environ['X509_USER_PROXY'] = old_proxy
                else:
                    del os.environ['X509_USER_PROXY']

        #AMIDataset
        if job.inputdata._name == 'AMIDataset':
            job.inputdata.dataset = job.inputdata.search()

        #EventPicking
        if job.inputdata._name == 'EventPicking':
            guid_run_evt_map = job.inputdata.get_pick_dataset()

        # before we do anything, check for tag info for this dataset
        additional_datasets = {}
        local_tag = False
        grid_tag = False

        # check for complete TAG mapping between files
        if job.inputdata.tag_info:

            # check for conflicts with TAG_LOCAL or TAG_COPY
            if job.inputdata.type in ['TAG_LOCAL', 'TAG_COPY']:
                raise ApplicationConfigurationError(None, "Cannot provide both tag_info and run as '%s'. Please use one or the other!" % job.inputdata.type)
            
            logger.warning('TAG information present - overwriting previous DQ2Dataset definitions')

            job.inputdata.names = []
            job.inputdata.dataset = []

            # check if FILE_STAGER is used
            if job.inputdata.type == 'FILE_STAGER':
                logger.warning("TAG jobs currently can't use the FILE_STAGER. Switching to DQ2_COPY instead.")
                job.inputdata.type = 'DQ2_COPY'

            # deal with tag_info depending on backend            
            if job.backend._name == 'Panda':

                # construct a reverse tag map going from AOD -> TAG
                grid_tag = True
                rev_tag_map = {}
                for tag_file in job.inputdata.tag_info:

                    for tag_ref in job.inputdata.tag_info[tag_file]['refs']:
                        if not tag_ref[0] in rev_tag_map:
                            rev_tag_map[tag_ref[0]] = {}
                            rev_tag_map[tag_ref[0]]['dataset'] = tag_ref[1]
                            rev_tag_map[tag_ref[0]]['guid'] = tag_ref[2]
                            rev_tag_map[tag_ref[0]]['refs'] = []

                        rev_tag_map[tag_ref[0]]['refs'].append([tag_file, job.inputdata.tag_info[tag_file]['dataset'],job.inputdata.tag_info[tag_file]['guid']])


                # construct the tag dataset map for brokering later
                tag_dataset_map = {}
                for tag_ref in rev_tag_map:
                    if not rev_tag_map[tag_ref]['dataset'] in job.inputdata.dataset:
                        job.inputdata.dataset.append(rev_tag_map[tag_ref]['dataset'])
                        tag_dataset_map[ rev_tag_map[tag_ref]['dataset'] ] = []
                        for tag_ref2 in rev_tag_map[tag_ref]['refs']:
                            if not tag_ref2[1] in tag_dataset_map[ rev_tag_map[tag_ref]['dataset'] ]:
                                tag_dataset_map[ rev_tag_map[tag_ref]['dataset'] ].append( tag_ref2[1] )
                        
                    if not tag_ref in job.inputdata.names:
                        job.inputdata.names.append(tag_ref)

            else:                
                # assemble the tag datasets to split over
                for tag_file in job.inputdata.tag_info:

                    if job.inputdata.tag_info[tag_file]['dataset'] != '' and job.inputdata.tag_info[tag_file]['path'] == '':
                        grid_tag = True

                        job.inputdata.names.append( tag_file )

                        if not job.inputdata.tag_info[tag_file]['dataset'] in job.inputdata.dataset:
                            job.inputdata.dataset.append(job.inputdata.tag_info[tag_file]['dataset'])

                        # add to additional datasets list
                        for tag_ref in job.inputdata.tag_info[tag_file]['refs']:
                            if job.inputdata.tag_info[tag_file]['dataset'] not in additional_datasets:
                                additional_datasets[job.inputdata.tag_info[tag_file]['dataset']] = []

                            if not tag_ref[1] in additional_datasets[job.inputdata.tag_info[tag_file]['dataset']]:
                                additional_datasets[job.inputdata.tag_info[tag_file]['dataset']].append(tag_ref[1])

                    elif job.inputdata.tag_info[tag_file]['path'] != '' and job.inputdata.tag_info[tag_file]['dataset'] == '':
                        local_tag = True
                        if not job.inputdata.tag_info[tag_file]['refs'][0][1] in job.inputdata.dataset:
                            job.inputdata.dataset.append(job.inputdata.tag_info[tag_file]['refs'][0][1])      

                        ## add the referenced files and guids and check for multiple datasets per file
                        ref_dataset = job.inputdata.tag_info[tag_file]['refs'][0][1]
                        for ref in job.inputdata.tag_info[tag_file]['refs']:
                            if ref[1] != ref_dataset:
                                raise ApplicationConfigurationError(None,'Problems with TAG entry for %s. Multiple datasets referenced for local TAG file.' % tag_file)

                            job.inputdata.names.append( ref[0] )
                            job.inputdata.guids.append( ref[2] )

                    else:
                        raise ApplicationConfigurationError(None,'Problems with TAG entry for %s' % tag_file)
                
            if grid_tag and local_tag:
                raise ApplicationConfigurationError(None,'Problems with TAG info - both grid and local TAG files selected.')
                    
                    
        # now carry on as before
        orig_numfiles = self.numfiles
        orig_numsubjobs = self.numsubjobs

        if self.numfiles <= 0: 
            self.numfiles = 1

        # use a key of the whole inDS structure for cache
        indata_buf = StringIO.StringIO()
        job.inputdata.printTree(indata_buf)
        locations = job.inputdata.get_locations(overlap=False)

        allowed_sites = []
        if job.backend._name in [ 'LCG', 'CREAM' ]:
            if job.backend.requirements._name in [ 'AtlasLCGRequirements', 'AtlasCREAMRequirements']:
                if job.backend.requirements.sites:
                    allowed_sites = job.backend.requirements.sites
                elif job.backend.requirements.cloud:

                    from GangaAtlas.Lib.Athena.AthenaLCGRTHandler import getLCGReleaseTag
                    
                    rel_tag = getLCGReleaseTag( job.application )
                    rel_tag_str = ''
                    if len(rel_tag) > 0:
                        rel_tag_str = rel_tag[0]
                    
                    if job.backend.requirements.cloud == 'ALL' and not job.backend.requirements.sites and job.outputdata and job.outputdata._name == 'DQ2OutputDataset':
                        logger.warning('DQ2OutputDataset being used with \'ALL\' cloud option. Attempting to find suitable sites - this may take some time...')
                        allowed_sites = job.backend.requirements.list_sites( req_str = rel_tag_str )
                    else:
                        allowed_sites = job.backend.requirements.list_sites_cloud( req_str = rel_tag_str )

                    if len(allowed_sites) == 0:
                        raise ApplicationConfigurationError(None,'DQ2JobSplitter could not find any allowed sites. This could be due to blacklisting, not having the required release installed or sites/clouds being manually excluded.')
                else: 
                    raise ApplicationConfigurationError(None,'DQ2JobSplitter requires a cloud or a site to be set - please use the --cloud option, j.backend.requirements.cloud=CLOUDNAME (T0, IT, ES, FR, UK, DE, NL, TW, CA, US, NG) or j.backend.requirements.sites=SITENAME')
                allowed_sites_all = job.backend.requirements.list_sites(True,True)
                # Apply GangaRobot blacklist
                if self.use_blacklist:
                    newsites = []
                    for site in allowed_sites:
                        if site in allowed_sites_all:
                            newsites.append(site)
                    allowed_sites = newsites
                # Check atlas_dbrelease
                if hasattr(job.application,"atlas_dbrelease") and job.application.atlas_dbrelease:
                    if job.application.atlas_dbrelease == 'LATEST':       
                        from pandatools import Client
                        my_atlas_dbrelease = Client.getLatestDBRelease(False)
                        job.application.atlas_dbrelease = my_atlas_dbrelease
                        
                        match = re.search('DBRelease-(.*)\.tar\.gz', my_atlas_dbrelease )
                        if match:
                            dbvers = match.group(1)
                            job.application.atlas_environment+=['DBRELEASE_OVERRIDE=%s'%dbvers] 

                    try:
                        db_dataset = job.application.atlas_dbrelease.split(':')[0]
                    except:
                        raise ApplicationConfigurationError(None,'Problem in DQ2JobSplitter - j.application.atlas_dbrelease is wrongly configured ! ')
                    from dq2.clientapi.DQ2 import DQ2
                    from dq2.info import TiersOfATLAS
                    dq2=DQ2(force_backend='rucio')
                    try:
                        db_locations = dq2.listDatasetReplicas(db_dataset).values()[0][1]
                    except:
                        raise ApplicationConfigurationError(None,'Problem in DQ2JobSplitter - j.application.atlas_dbrelease is wrongly configured ! ')

                    # Update allowed_sites to contain all possible spacetokens of a site
                    dq2alternatenames = []
                    allowed_sites_all = []
                    for site in allowed_sites:
                        alternatename = TiersOfATLAS.getSiteProperty(site,'alternateName')
                        if not alternatename in dq2alternatenames:
                            dq2alternatenames.append(alternatename)
                        else:
                            continue
                        for sitename in TiersOfATLAS.getAllSources():
                            if TiersOfATLAS.getSiteProperty(sitename,'alternateName') == alternatename and not sitename in allowed_sites_all: 
                                allowed_sites_all.append(sitename)

                    allowed_sites = allowed_sites_all

                    db_allowed_sites=[]
                    dq2alternatenames=[]
                    for site in allowed_sites:
                        if site in db_locations:
                            db_allowed_sites.append(site)
                            dq2alternatenames.append(TiersOfATLAS.getSiteProperty(site,'alternateName'))
                    for sitename in TiersOfATLAS.getAllSources():
                        if TiersOfATLAS.getSiteProperty(sitename,'alternateName'):
                            if TiersOfATLAS.getSiteProperty(sitename,'alternateName') in dq2alternatenames and not sitename in db_allowed_sites:
                                db_allowed_sites.append(sitename)
                    allowed_sites = db_allowed_sites
                # Check for additional datasets
                if len(additional_datasets) > 0:
                    from dq2.clientapi.DQ2 import DQ2
                    from dq2.info import TiersOfATLAS
                    dq2=DQ2(force_backend='rucio')
                    add_locations_all = []

                    additional_datasets_all = []
                    
                    for tag_dataset in additional_datasets:
                        for add_dataset in additional_datasets[tag_dataset]:
                            if not add_dataset in additional_datasets_all:
                                additional_datasets_all.append(add_dataset)
                            
                    for add_dataset in additional_datasets_all:
                        if len(add_locations_all) == 0:
                            add_locations_all = dq2.listDatasetReplicas(add_dataset).values()[0][1]
                        else:
                            add_locations = dq2.listDatasetReplicas(add_dataset).values()[0][1]
                            
                            for add_location in add_locations_all:
                                if not add_location in add_locations:
                                    add_locations_all.remove(add_location)
                                    
                    add_allowed_sites=[]
                    dq2alternatenames=[]
                    for site in allowed_sites:
                        if site in add_locations_all:
                            add_allowed_sites.append(site)
                            dq2alternatenames.append(TiersOfATLAS.getSiteProperty(site,'alternateName'))
                    for sitename in TiersOfATLAS.getAllSources():
                        if TiersOfATLAS.getSiteProperty(sitename,'alternateName'):
                            if TiersOfATLAS.getSiteProperty(sitename,'alternateName') in dq2alternatenames and not sitename in add_allowed_sites:
                                add_allowed_sites.append(sitename)
                    allowed_sites = add_allowed_sites

                # Check if site is online:
                newsites = []
                for asite in allowed_sites:
                    if not job.backend.requirements.list_ce(asite):
                        logger.warning('Site %s is currently down or does not allow user analysis - please check carefully if all inputfiles are available for your jobs. Maybe switch to a different cloud.',asite)
                    else:
                        newsites.append(asite)
                allowed_sites = newsites
                    
        elif job.backend._name == 'Panda':
            allowed_sites = job.backend.list_ddm_sites()
            if self.use_fax:
                faxSites = allowed_sites
        elif job.backend._name == 'NG':
            allowed_sites = config['AllowedSitesNGDQ2JobSplitter']
        elif job.backend._name == 'SGE':
            from dq2.clientapi.DQ2 import DQ2
            from dq2.info import TiersOfATLAS
            allowed_sites = TiersOfATLAS.getAllSources()
            
        if not allowed_sites:
            raise ApplicationConfigurationError(None,'DQ2JobSplitter found no allowed_sites for dataset')

        if 'LRZ-LMU_DATADISK' in allowed_sites:
            allowed_sites.append('LRZ-LMU-RUCIOTEST_DATADISK')
        if 'PRAGUELCG2_DATADISK' in allowed_sites:
            allowed_sites.append('PRAGUELCG2-RUCIOTEST_DATADISK')

        logger.debug('allowed_sites = %s ', allowed_sites)

        contents_temp = job.inputdata.get_contents(overlap=False, size=True)

        if self.numevtsperjob > 0:
            contents_temp = job.inputdata.get_contents(overlap=False, event=True)
        else:
            contents_temp = job.inputdata.get_contents(overlap=False, size=True)

        logger.debug(contents_temp)

        contents = {}
        datasetLength = {}
        datasetFilesize = {}
        nevents = {}
        allfiles = 0
        allsizes = 0
        allevents = 0
        allnames = []
        for dataset, content in contents_temp.iteritems():
            if not content:
                continue
            contents[dataset] = content
            datasetLength[dataset] = len(contents[dataset])
            allfiles += datasetLength[dataset]
            datasetFilesize[dataset] = reduce(operator.add, map(lambda x: x[1][1],content))
            allsizes += datasetFilesize[dataset]
            allnames += map(lambda x: x[1][0],content)
            if self.numevtsperjob > 0:
                nevents[dataset] = reduce(operator.add, map(lambda x: x[1][2],content))
                allevents += nevents[dataset]
                logger.info('Dataset %s contains %d events in %d files '%(dataset, nevents[dataset], datasetLength[dataset]))
            else:
                logger.info('Dataset %s contains %d files in %d bytes'%(dataset,datasetLength[dataset],datasetFilesize[dataset]))
        logger.info('Total num files=%d, total file size=%d bytes'%(allfiles,allsizes))

        # to a check for the 'ALL' cloud option and if given, reduce the selection
        if hasattr(job.backend, 'requirements') and hasattr(job.backend.requirements, 'cloud') and job.backend.requirements.cloud == 'ALL' and not job.backend.requirements.sites and job.outputdata and job.outputdata._name == 'DQ2OutputDataset':

            if not job.backend.requirements.anyCloud:
            
                logger.warning('DQ2OutputDataset being used with \'ALL\' cloud option. Restricting to a single cloud. Set the \'requirements.anyCloud\' to True to prevent this restriction.')

                avail_clouds = {}
                for key in locations:
                    avail_clouds[key] = []

                    new_locations = []
                    for loc in locations[key]:
                        if loc in allowed_sites:
                            new_locations.append(loc)
                        
                    info = job.backend.requirements.cloud_from_sites(new_locations)
                
                    for all_site in info:
                        if not info[all_site] in avail_clouds[key] and not info[all_site] in ['US', 'NG']:
                            if info[all_site] == 'T0':
                                info[all_site] = 'CERN'                     
                            avail_clouds[key].append(info[all_site])

                # perform logical AND to find a cloud that has all data
                import sys
            
                cloud_set = set(job.backend.requirements.list_clouds())
            
                for key in avail_clouds:
                    cloud_set = cloud_set & set(avail_clouds[key])

                # find users cloud and submit there by preference
                fav_cloud = ''
                for ln in gridProxy.info('--all').split('\n'):
                    if ln.find('attribute') == -1 or ln.find('atlas') == -1:
                        continue
                
                    toks = ln.split('/')
                            
                    if len(toks) < 3:
                        continue
                
                    if toks[2].upper() in job.backend.requirements.list_clouds():
                        fav_cloud = toks[2].upper()
                        break
                            
                if len(cloud_set) == 0:
                    raise ApplicationConfigurationError(None, 'Cloud option \'ALL\' could not find a complete replica of the dataset in any cloud. Please try a specific site or cloud.')
                else:
                    cloud_list = list(cloud_set)
                    if not fav_cloud in cloud_list:
                        fav_cloud = cloud_list[0]

                    logger.warning('\'%s\' cloud selected. Continuing job submission...' % fav_cloud)

            new_allowed_sites = []
            for site in allowed_sites:
                if not job.backend.requirements.anyCloud:
                    if job.backend.requirements.cloud_from_sites( site )[site] == fav_cloud:
                        new_allowed_sites.append(site)
                elif not job.backend.requirements.cloud_from_sites( site )[site] in ['US', 'NG']:
                    new_allowed_sites.append(site)

            allowed_sites = new_allowed_sites
                

        siteinfos = {}
        allcontents = {}
        if job.inputdata._name == 'EventPicking' and job.backend._name == 'Panda':
            if (job.inputdata.pick_filter_policy == 'reject'):
                raise ApplicationConfigurationError(None,"Pick event filter policy 'reject' not supported on Panda backend.")
            # create a file containing list of files
            test_area = os.environ['TestArea']
            eventPickFileList = '%s/epFileList_%s.dat' % (test_area, commands.getoutput('uuidgen'))
            evFileList = open(eventPickFileList,'w') 

        for dataset, content in contents.iteritems():
            
            content = dict(content)
            if self.use_lfc:
                logger.warning('Please be patient - scanning LFC catalogs ...')
                result = job.inputdata.get_replica_listing(dataset,SURL=False,complete=-1)
                siteinfo = lfc_siteinfo(result, allowed_sites)
            else:
                if self.update_siteindex:
                    udays = 2
                    skipReplicaLookup = False
                else:
                    udays = 10000
                    skipReplicaLookup = True
                if locations and dataset in locations:
                    siteinfo = dq2_siteinfo(dataset, allowed_sites, locations[dataset], udays, faxSites, skipReplicaLookup)
                else:
                    siteinfo = {}

            siteinfos[dataset]=siteinfo
            allcontents[dataset]=content

        logger.debug('siteinfos = %s', siteinfos)
        logger.debug('allcontents = %s', allcontents)

        subjobs = []
        totalfiles = 0
        events_processed = 0
        
        #print "%10s %20s %10s %10s %10s %10s %10s %10s %10s "  %("nrjob", "guid", "nevents", "skip_events", "max_events", "unused_evts", "id_lower", "id_upper", "counter")

        for dataset, siteinfo in siteinfos.iteritems():
            logger.debug('dataset [%s] siteinfo [%s]' % (dataset, siteinfo))

            self.numfiles = orig_numfiles
            self.numsubjobs = orig_numsubjobs
            if self.numfiles <= 0: 
                self.numfiles = 1

            for sites, guids in siteinfo.iteritems():

                # check for TAG datasets at these sites
                tag_dset_size = 0
                if len(tag_dataset_map) > 0:
                    from dq2.clientapi.DQ2 import DQ2
                    from dq2.info import TiersOfATLAS
                    dq2=DQ2(force_backend='rucio')

                    logger.warning("Parent dataset %s being used with TAG dataset(s) %s. Brokering now..." % (dataset, tag_dataset_map[dataset] ))

                    for tagDS in tag_dataset_map[dataset]:

                        if not job.inputdata.use_cvmfs_tag:
                            tag_locations = dq2.listDatasetReplicas(tagDS).values()[0][1]
                            new_sites = []
                            dq2alternatenames = []
                            for site in sites.split(':'):
                                if site in tag_locations:
                                    new_sites.append(site)
                                    dq2alternatenames.append(TiersOfATLAS.getSiteProperty(site,'alternateName'))
                            for sitename in TiersOfATLAS.getAllSources():
                                if TiersOfATLAS.getSiteProperty(sitename,'alternateName'):
                                    if TiersOfATLAS.getSiteProperty(sitename,'alternateName') in dq2alternatenames and not sitename in new_sites:
                                        new_sites.append(sitename)
                            sites = ':'.join(new_sites)

                        if job.inputdata.tagdataset:
                            tag_contents = job.inputdata.get_tag_contents(size=True, spec_dataset = tagDS)
                            tag_dset_size += reduce(operator.add, map(lambda x: x[1][1],tag_contents))

                # preferentially select sites given the cloud priority
                cloud_pref = config['AnyCloudPreferenceList']
                if len(cloud_pref) > 0:
                    clouds = {}
                    # use AtlasLCGRequirements for LCG and PandaTools for Panda to find mapping from site to cloud
                    if job.backend._name in ['LCG'] and job.backend.requirements.anyCloud:
                        clouds = job.backend.requirements.cloud_from_sites(sites.split(':'))
                    elif job.backend._name in ['Panda'] and job.backend.requirements.anyCloud:
                        from pandatools import Client
                        for site in sites.split(':'):
                            queue = Client.convertDQ2toPandaID(site)
                            clouds[site] = Client.PandaSites[queue]['cloud']
                            
                    # Now try to match this with the cloud preferences
                    for cl in cloud_pref:
                        if cl in clouds.values():
                            # cloud preference found - remove all but these sites
                            new_sites = []
                            for site in clouds:
                                if clouds[site] == cl:
                                    new_sites.append( site )
                            sites = ':'.join(new_sites)
                            break
                        
                # at these sites process these guids belonging to dataset
                
                counter = 0 ; id_lower = 0;  id_upper = 0; tmpid = 0
                unused_events = 0; nskip =0; second_loop =0; nevtstoskip = 0;  
                totalevent = 0; used_events = 0; first_loop = 0
                left_events = 0; first_iteration = False; last_loop = False; num_of_events = 0;
                dset_size = 0; nsubjob = 0

                if self.numfiles <= 0: 
                    self.numfiles = 1

                allcontent = allcontents[dataset]

                # Fix bug 42044
                # drop unused guids
                removal = []
                for g in guids:
                    if not g in allcontent.keys():
                        logger.debug("Removing guid %s" % g)
                        removal += [g]

                for g in removal:
                    guids.remove(g)

                if len(guids) == 0:
                    continue

                max_subjob_filesize = 0
                # do a Panda brokerage for these sites
                pandaSite = ''
                if job.backend._name == 'Panda':
                    from GangaPanda.Lib.Panda.Panda import selectPandaSite
                    pandaSite = selectPandaSite(job,sites)
                    logger.debug('pandaSite = %s' % pandaSite)
                    # set maximum input size per subjob 
                    from pandatools import Client
                    try:
                        max_subjob_filesize = (Client.PandaSites[pandaSite]['maxinputsize']-1000)*1024*1024
                    except:
                        logger.debug('max_subjob_filesize not set !')
                        pass
                    if not max_subjob_filesize:
                        max_subjob_filesize = config['MaxFileSizePandaDQ2JobSplitter']*1024*1024

                    logger.debug('max_subjob_filesize = %s' % max_subjob_filesize)
                    # direct access site ?
                    from pandatools import PsubUtils
                    # FIXME - set correct parameter values
                    inBS = False
                    inTRF = False
                    try:
                        if job.application.atlas_exetype == 'TRF':
                            inTRF = True
                    except:
                        pass
                    if job.backend.accessmode == 'DIRECT':
                        inTRF=False
                    inARA = False
                    isDirectAccess = PsubUtils.isDirectAccess(pandaSite, inBS, inTRF, inARA)
                    logger.debug('isDirectAccess = %s' % isDirectAccess)
                    if isDirectAccess or job.backend.accessmode in ['DIRECT','FILE_STAGER']: 
                       max_subjob_filesize = 0
                       # use numfiles if user has set it, else use MaxSubjobFilesPandaDQ2JobSplitter
                       if not orig_numsubjobs>0 and not orig_numfiles>0:
                           self.numfiles = config['DefaultNumFilesPandaDirectDQ2JobSplitter']
                           self.numsubjobs = 0
                           logger.info('DQ2JobSplitter has restricted the maximum of files per subjob to %s files.', self.numfiles)
                       else:
                           if orig_numsubjobs>0:
                               nrfiles = int(math.ceil(len(guids)/float(orig_numsubjobs)))
                               if nrfiles > config['DefaultNumFilesPandaDirectDQ2JobSplitter']:
                                   self.numfiles = config['DefaultNumFilesPandaDirectDQ2JobSplitter']
                                   self.numsubjobs = 0
                           if orig_numfiles>config['DefaultNumFilesPandaDirectDQ2JobSplitter']:
                               self.numfiles = config['DefaultNumFilesPandaDirectDQ2JobSplitter']
                               self.numsubjobs = 0
                           logger.info('DQ2JobSplitter has restricted the maximum of files per subjob to %s files.', self.numfiles)

                # Restriction based on the maximum dataset filesize
                if self.filesize > 0:
                    max_subjob_filesize = self.filesize*1024*1024
                elif job.backend._name == 'NG' and (self.filesize < 1 or config['MaxFileSizeNGDQ2JobSplitter'] < self.filesize):
                    max_subjob_filesize = config['MaxFileSizeNGDQ2JobSplitter']*1024*1024
                elif job.backend._name == 'Panda':
                    if max_subjob_filesize > 0:
                    #max_subjob_filesize = config['MaxFileSizePandaDQ2JobSplitter']*1024*1024
                        logger.info('DQ2JobSplitter has restricted the maximum inputsize per subjob to %s Bytes.', max_subjob_filesize)
                    else:
                        max_subjob_filesize = 1180591620717411303424 # 1 zettabyte
                else:
                    max_subjob_filesize = 1180591620717411303424 # 1 zettabyte

                nrfiles = self.numfiles
                nrjob = int(math.ceil(len(guids)/float(nrfiles)))
                if nrjob > self.numsubjobs and self.numsubjobs!=0:
                    nrfiles = int(math.ceil(len(guids)/float(self.numsubjobs)))
                    nrjob = int(math.ceil(len(guids)/float(nrfiles)))
                elif self.numevtsperjob > 0:
                    for g in guids:
                        totalevent += allcontent[g][2]
                        dset_size += allcontent[g][1]
                    nrjob = int(math.ceil(totalevent/float(self.numevtsperjob)))
                   
                    # Restriction based on the maximum dataset filesize
                    filesize_per_event = dset_size/totalevent 
                    filesize_per_subjob = filesize_per_event * self.numevtsperjob
                    if filesize_per_subjob > max_subjob_filesize :
                        events_per_subjob = max_subjob_filesize/filesize_per_event 
                        self.numevtsperjob = events_per_subjob
                        nrjob = int(math.ceil(totalevent/float(events_per_subjob)))
                                    
                # filesize based splitting
                if max_subjob_filesize and self.filesize > 0:
                    nrjob = config['MaxJobsDQ2JobSplitter']
                    nrfiles = len(guids)

                # split on local tag files if required
                if job.inputdata.tag_info and local_tag and nrjob > len(job.inputdata.tag_info):
                    nrfiles = int(math.ceil(len(guids)/float(len(job.inputdata.tag_info))))
                    nrjob = len(job.inputdata.tag_info)

                if nrjob > config['MaxJobsDQ2JobSplitter']: 
                    if self.numevtsperjob > 0:
                        self.numevtsperjob = int(math.ceil(totalevent/float((config['MaxJobsDQ2JobSplitter'] -1))))
                        nrjob = int(math.ceil(totalevent/float(self.numevtsperjob)))
                    else:
                        logger.warning('!!! Number of subjobs %s is larger than maximum allowed of %s - reducing to %s !!!', nrjob, config['MaxJobsDQ2JobSplitter'], config['MaxJobsDQ2JobSplitter'] )
                        nrfiles = int(math.ceil(len(guids)/float(config['MaxJobsDQ2JobSplitter'])))
                        nrjob = int(math.ceil(len(guids)/float(nrfiles)))
                        
                # Restrict number of subjobs is compilation is used
                if job.backend._name in ['LCG', 'CREAM' ] and job.application.athena_compile==True and  nrjob > config['MaxJobsDQ2JobSplitterLCGCompile']: 
                    logger.error('!!! The number of allowed subjobs on the %s backend and with athena_compile=True is %s, but DQ2JobSplitter is trying to split into %s subjobs !!!', job.backend._name, config['MaxJobsDQ2JobSplitterLCGCompile'], nrjob )
                    logger.error('!!! Please pre-compile the code locally and use athena_compile=False in a new job submission if more subjobs are required !!!' )
                    raise ApplicationConfigurationError(None,'!!! Submission stopped !!!')

                if nrfiles > len(guids):
                    nrfiles = len(guids)

                # sort the guids by name order
                names = [allcontent[g][0] for g in guids]
                namesAndGuids = zip(names,guids)
                namesAndGuids.sort()
                names,guids = zip(*namesAndGuids)

                # now assign the files to subjobs
                max_subjob_numfiles = nrfiles
                if self.numevtsperjob > 0:
                    logger.info('DQ2JobSplitter will attempt to create %d subjobs using  %d events per subjob subject to a limit of %d Bytes per subjob.' %(nrjob,self.numevtsperjob, max_subjob_filesize))
                elif max_subjob_filesize and  self.filesize > 0:
                    #logger.info('DQ2JobSplitter will attempt to create %d subjobs using %d files per subjob subject to a limit of %d Bytes per subjob.'%(nrjob,max_subjob_numfiles,max_subjob_filesize))
                    pass
                remaining_guids = list(guids)

                # sort out the tag files that reference this if required
                if job.inputdata.tag_info and local_tag:
                    
                    remaining_tags = []   # c.f. remaining guids: list of tag files that ref. this dataset

                    for tag_file in job.inputdata.tag_info:

                        # check all guids are referenced
                        num = 0                            
                        for ref in job.inputdata.tag_info[tag_file]['refs']:                
                            if ref[2] in remaining_guids:
                                num += 1

                        if num > 0 and num != len(job.inputdata.tag_info[tag_file]['refs']):
                            for ref in job.inputdata.tag_info[tag_file]['refs']:
                                if ref[2] in remaining_guids:
                                    remaining_guids.remove(ref[2])
                            logger.warning('Not all guids referenced by TAG file %s were found in available sites - ignoring this file.' % tag_file)
                        elif num > 0:
                            remaining_tags.append( tag_file )
                            
                while remaining_guids and len(subjobs)<config['MaxJobsDQ2JobSplitter']:
                    num_remaining_guids = len(remaining_guids)
                    j = Job()
                    j.name = job.name
                    j.inputdata = job.inputdata
                    j.inputdata.dataset = dataset
                    j.inputdata.sizes = []
                    j.inputdata.guids = []
                    j.inputdata.names = []
                    j.inputdata.checksums = []
                    j.inputdata.scopes = []
                    j.application   = job.application
                    j.application.run_event   = []

                    if len(tag_dataset_map) > 0 and not job.inputdata.tag_info:
                        j.inputdata.tagdataset = tag_dataset_map[dataset]
                        
                    if self.numevtsperjob > 0:
                        if  unused_events == 0:
                            nevtstoskip = 0
                            id_lower = id_upper
                        else:
                            previous_guid = guids[id_lower] 
                            nevtstoskip = allcontent[previous_guid][2]- left_events
                        
                        #add events: first loop 
                        while (remaining_guids and not last_loop)  and  self.numevtsperjob > unused_events:
                            next_guid = guids[counter]
                            unused_events += allcontent[next_guid][2]
                            if counter < len(guids):
                                id_upper +=1
                            if counter < (len(guids) -1) :
                                remaining_guids.remove(next_guid)
                            if counter == (len(guids) -1) :
                                last_loop = True
                            counter += 1
                            first_loop = 1
                        
                        #use events: second loop
                        while (self.numevtsperjob <= unused_events) or last_loop:
                            if nsubjob == 0:
                                nevtstoskip = 0
                            elif first_loop != 1:
                                previous_guid = guids[id_lower] 
                                nevtstoskip = allcontent[previous_guid][2]- unused_events
                            first_loop = 0
                            if (nsubjob != nrjob -1):
                                unused_events -= self.numevtsperjob
                                left_events = unused_events
                            second_loop = 1
                            num_of_events = self.numevtsperjob
                            if ( nsubjob == nrjob - 1 ):
                                last_loop = False
                                last_guid = guids[counter - 1]
                                remaining_guids.remove(last_guid)
                                num_of_events = unused_events
                                unused_events = 0
                            break
                        
                        j.inputdata.guids = list(guids[id_lower:id_upper])
                        j.inputdata.names = [allcontent[g][0] for g in j.inputdata.guids]
                        j.inputdata.sizes = [allcontent[g][1] for g in j.inputdata.guids]
                        j.inputdata.checksums = [allcontent[g][2] for g in j.inputdata.guids]
                        j.inputdata.scopes = [allcontent[g][3] for g in j.inputdata.guids]
                        j.application.skip_events = nevtstoskip
                        j.application.max_events = num_of_events
                        events_processed += num_of_events
                        nsubjob += 1
                        
                        """
                        print "%10s %20s %10s %10s %10s %10s %10s %10s %10s "  %("nrjob", "guid", "nevents", "skip_events", "max_events", "unused_evts", "id_lower", "id_upper", "counter")
                        print "\n%10s %20s %10s %10s %10s %10s %10s %10s %10s "  %( nsubjob, (j.inputdata.names[0])[-18:], allcontent[j.inputdata.guids[0]][2] - j.application.skip_events, j.application.skip_events, j.application.max_events, unused_events, id_lower, id_upper, counter - 1)
                        
                        kcnt = 0
                        kevent = 0
                        for guid in j.inputdata.guids:
                            if (kcnt != 0):
                                print "%10s %20s %10s"  %("", allcontent[guid][0][-18:], allcontent[guid][2]) 
                                kevent += allcontent[guid][2]
                            else:
                                kevent += allcontent[guid][2] - j.application.skip_events
                            kcnt +=1
                        if len(j.inputdata.guids) > 1:
                            print "%10s %20s %10s" %("", "total_events", kevent)

                        """

                        #Remove previously added files from the subjobs         
                        id_lower = id_upper -1
                    
                    else:
                        # change the splitting based on local tag files
                        if job.inputdata.tag_info and local_tag:
                            while remaining_tags and len(j.inputdata.guids)<max_subjob_numfiles and sum(j.inputdata.sizes)<max_subjob_filesize:
                                for next_tag in remaining_tags:
                                    
                                    remaining_tags.remove(next_tag)
                                    
                                    for ref in job.inputdata.tag_info[next_tag]['refs']:
                                        if ref[2] in remaining_guids:
                                            remaining_guids.remove(ref[2])
                                        else:
                                            logger.warning("Multiple TAG files referenced GUID %s - this could lead to unexpected results!" % ref[2])
                                                                                                
                                        j.inputdata.guids.append(ref[2])
                                        j.inputdata.names.append(allcontent[ref[2]][0])
                                        j.inputdata.sizes.append(allcontent[ref[2]][1])

                                    break
                            
                                else:
                                    break
                        else:
                            # default splitting
                            while remaining_guids and len(j.inputdata.guids)<max_subjob_numfiles and sum(j.inputdata.sizes)<max_subjob_filesize:

                                too_large_filesizes = []
                                for next_guid in remaining_guids:
                                    
                                    # check if files are bigger than the filesize requirement
                                    if allcontent[next_guid][1] > max_subjob_filesize:
                                        too_large_filesizes.append( allcontent[next_guid][1] / (1024*1024) )
                                    

                                    if sum(j.inputdata.sizes)+allcontent[next_guid][1] < max_subjob_filesize:
                                        remaining_guids.remove(next_guid)
                                        j.inputdata.guids.append(next_guid)
                                        j.inputdata.names.append(allcontent[next_guid][0])
                                        j.inputdata.sizes.append(allcontent[next_guid][1])
                                        j.inputdata.checksums.append(allcontent[next_guid][2])
                                        j.inputdata.scopes.append(allcontent[next_guid][3])
                                        if job.inputdata._name != 'EventPicking':
                                            break
                                    
                                    if job.inputdata._name == 'EventPicking':
                                        for runevent in guid_run_evt_map[next_guid]:
                                            revt = "(%s,%s)" %(long(runevent[0]),long(runevent[1]))     
                                            j.application.run_event.append([long(runevent[0]),long(runevent[1])])
                                        
                                        if job.backend._name == 'Panda':
                                            app = job.application
                                            # copy run/event list in a file
                                            eventPickFile = '%s/%s/ep_%s.dat' % (test_area,app.atlas_run_dir,commands.getoutput('uuidgen'))
                                            evFH = open(eventPickFile,'w') 
                                            tlist = j.application.run_event
                                            j.application.run_event_file = os.path.basename(eventPickFile)

                                            for k in range(len(tlist)):
                                                evFH.write('%s %s\n' % (tlist[k][0], tlist[k][1]))
                                            # close        
                                            evFH.close()
                                            #write in file 
                                            evFileList.write('%s%s\n' %(app.atlas_run_dir, os.path.basename(eventPickFile)))

                                else:
                                    break

                    
                    j.inputdata.number_of_files = len(j.inputdata.guids)
                    if (self.numevtsperjob == 0):
                        if not (job.inputdata.tag_info and local_tag) and num_remaining_guids == len(remaining_guids):
                            logger.warning('Filesize constraint (%d) is too small for some files and has therefore blocked the assignment of %d files having guids: %s. Filesizes are: %s'%(self.filesize, len(remaining_guids),remaining_guids, too_large_filesizes))
                            break
                    #print j.inputdata.names
                    #print j.inputdata.sizes
                    #print sum(j.inputdata.sizes)
                    j.outputdata    = job.outputdata
                    j.backend       = job.backend
                    if j.backend._name in [ 'LCG', 'CREAM']:
                        j.backend.requirements.sites = sites.split(':')
                    if j.backend._name == 'Panda':
                        j.backend.site=pandaSite
                    if j.backend._name == 'SGE':
                        for site in sites.split(':'):
                            if site.startswith('DESY-HH'):
                                j.backend.extraopts+=' -l site=hh '
                                break
                            elif site.startswith('DESY-ZN'):
                                j.backend.extraopts+=' -l site=zn '
                                break

                    j.inputsandbox  = job.inputsandbox
                    j.outputsandbox = job.outputsandbox 

                    # job inputdata type
                    #j.inputdata.type = job.inputdata.type

                    j.inputdata.tag_info = {}
                    if job.inputdata.tag_info:
                        if grid_tag:

                            if job.backend._name == 'Panda':
                                for tag_file in job.inputdata.tag_info:
                                    for tag_ref in job.inputdata.tag_info[tag_file]['refs']:
                                        if tag_ref[1] in j.inputdata.dataset and tag_ref[0] in j.inputdata.names:
                                            if not tag_file in j.inputdata.tag_info:
                                                j.inputdata.tag_info[tag_file] = {}                                                
                                                j.inputdata.tag_info[tag_file]['dataset'] = job.inputdata.tag_info[tag_file]['dataset']
                                                j.inputdata.tag_info[tag_file]['guid'] = job.inputdata.tag_info[tag_file]['guid']
                                                j.inputdata.tag_info[tag_file]['refs'] = []
                                                
                                            j.inputdata.tag_info[tag_file]['refs'].append( tag_ref )
                                    
                            else:
                                for tag_file in j.inputdata.names:
                                    j.inputdata.tag_info[tag_file] = job.inputdata.tag_info[tag_file]

                        if local_tag:
                            for tag_file in job.inputdata.tag_info:
                                if job.inputdata.tag_info[tag_file]['refs'][0][0] in j.inputdata.names:
                                    j.inputdata.tag_info[tag_file]  = job.inputdata.tag_info[tag_file]

                    subjobs.append(j)

                    totalfiles = totalfiles + len(j.inputdata.guids) 

        
        if job.inputdata._name == 'EventPicking' and job.backend._name == 'Panda':
            evFileList.close()
            #append list of files from picking to source archive
            appendFile(eventPickFileList, job.application.user_area.name)

        if not subjobs:
            logger.error('DQ2JobSplitter did not produce any subjobs! Either the dataset is not present in the cloud or at the site or all chosen sites are black-listed for the moment.')
            logger.error('job.inputdata.dataset %s job.backend.site %s ' % (job.inputdata.dataset, job.backend.site))
            raise ApplicationConfigurationError(None,'!!! Stopping submission now !!!')

        # reset missing files in case of a previous submission attempt
        self.missing_files = []

        if self.numevtsperjob > 0:
            logger.info('Total events assigned to subjobs is %d'%events_processed)
            if not (allevents == events_processed) :
                logger.error('DQ2JobSplitter was only able to process %s out of %s events to the subjobs ! Please check your job configuration if this is intended and possibly change to a different cloud or choose different sites!' %(events_processed, allevents) )
        else:
            logger.info('Total files assigned to subjobs is %d'%totalfiles)
            if not (totalfiles == allfiles):
                logger.error('DQ2JobSplitter was only able to assign %s out of %s files to the subjobs ! Please check your job configuration if this is intended and possibly change to a different cloud or choose different sites!', totalfiles, allfiles)
                
                for name in allnames:
                    found=False
                    for sj in subjobs:
                        if name in sj.inputdata.names:
                            found=True
                            break
                    if not found:
                        self.missing_files.append(name)
                logger.info('The files not assigned to jobs have been stored in job.splitter.missing_files')

        return subjobs
    
config = getConfig('Athena')



from __future__ import print_function
import time
import traceback
import sys
import Ganga.GPIDev.Lib.Registry.RegistrySlice
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySliceProxy
from Ganga.Core.GangaRepository.Registry import Registry, RegistryError, RegistryKeyError, RegistryAccessError, RegistryFlusher
from Ganga.GPIDev.Base.Proxy import stripProxy, getName, isType
from Ganga.Utility.ColourText import ANSIMarkup, overview_colours, status_colours, fgcol

from Ganga.Utility.logging import getLogger
logger = getLogger()

from Ganga.Utility.Config import getConfig
config = getConfig('Tasks')

markup = ANSIMarkup()
str_run = markup("run", overview_colours["running"])
str_fail = markup("fail", overview_colours["failed"])
str_hold = markup("hold", overview_colours["hold"])
str_bad = markup("bad", overview_colours["bad"])


class TaskRegistry(Registry):

    def __init__(self, name, doc):

        super(TaskRegistry, self).__init__( name, doc )

        self._main_thread = None

        self.stored_slice = TaskRegistrySlice(self.name)
        self.stored_slice.objects = self
        self.stored_proxy = TaskRegistrySliceProxy(self.stored_slice)

    def getSlice(self):
        return self.stored_slice

    def getProxy(self):
        return self.stored_proxy

    def getIndexCache(self, obj):
        cached_values = ['status', 'id', 'name']
        c = {}
        for cv in cached_values:
            if hasattr(obj, cv):
                c[cv] = getattr(obj, cv)
        this_slice = TaskRegistrySlice("tmp")
        for dpv in this_slice._display_columns:
            c["display:" + dpv] = this_slice._get_display_value(obj, dpv)
        return c

    def _thread_main(self):
        """ This is an internal function; the main loop of the background thread """
        from Ganga.Core.GangaRepository import getRegistry
        while getRegistry("jobs").hasStarted() is not True:
            time.sleep(0.1)
            if self._main_thread is None or self._main_thread.should_stop():
                return

        while True:
            from Ganga.Core import monitoring_component
            if (not monitoring_component is None and monitoring_component.enabled) or config['ForceTaskMonitoring']:
                break
            time.sleep(0.1)
            if self._main_thread is None or self._main_thread.should_stop():
                return

        # setup the tasks - THIS IS INCOMPATIBLE WITH CONCURRENCY
        # and must go away soon
        for tid in self.ids():
            try:
                self[tid].startup()
            except Exception as err:
                logger.error("Unknown/Unexpected Error in starting up tasks main loop")
                logger.error("Exiting: err=%s" % str(err))
                return

        logger.debug("Entering main loop")

        # Main loop
        while self._main_thread is not None and not self._main_thread.should_stop():

            # If monitoring is enabled (or forced for Tasks) loop over each one and update
            if (config['ForceTaskMonitoring'] or monitoring_component.enabled) and not config['disableTaskMon']:
                for tid in self.ids():

                    logger.debug("Running over tid: %s" % str(tid))

                    try:
                        p = self[tid]
                        p.update()

                    except Exception as x:
                        logger.error(
                            "Exception occurred in task monitoring loop: %s %s\nThe offending task was paused." % (x.__class__, x))
                        type_, value_, traceback_ = sys.exc_info()
                        logger.error("Full traceback:\n %s" % ' '.join(
                            traceback.format_exception(type_, value_, traceback_)))
                        p.pause()

                    if self._main_thread.should_stop():
                        break

                if self._main_thread.should_stop():
                    break

            logger.debug("TaskRegistry Sleeping for: %s seconds" % str(config['TaskLoopFrequency']))

            # Sleep interruptible for 10 seconds
            for i in range(0, int(config['TaskLoopFrequency'] * 100)):
                if self._main_thread.should_stop():
                    break
                time.sleep(0.01)

    def startup(self):
        """ Start a background thread that periodically run()s"""
        super(TaskRegistry, self).startup()
        from Ganga.Core.GangaThread import GangaThread
        self._main_thread = GangaThread(name="GangaTasks", target=self._thread_main)
        self._main_thread.start()

        # create a registry flusher
        self.flush_thread = RegistryFlusher(self)
        self.flush_thread.start()

    def shutdown(self):
        self.flush_thread.join()
        super(TaskRegistry, self).shutdown()

    def stop(self):
        if self._main_thread is not None:
            self._main_thread.stop()
            self._main_thread.join()

from Ganga.GPIDev.Lib.Registry.RegistrySlice import RegistrySlice


class TaskRegistrySlice(RegistrySlice):

    def __init__(self, name):
        super(TaskRegistrySlice, self).__init__(name, display_prefix="tasks")
        from Ganga.Utility.ColourText import Foreground, Background, Effects
        fg = Foreground()
        fx = Effects()
        bg = Background()
        self.status_colours = {'new': fx.normal,
                               'submitted': fg.orange,
                               'running': fg.green,
                               'completed': fg.blue,
                               'failed': fg.red}
        self.fx = fx
        self._proxyClass = TaskRegistrySliceProxy

    def _getColour(self, obj):
        return self.status_colours.get(obj.status, self.fx.normal)

    def __getitem__(self, id):
        return self.objects[id]

    def __call__(self, id):
        """ Retrieve a job by id.
        """
        t = type(id)
        if t is int:
            try:
                return self.objects[id]
            except KeyError:
                raise RegistryKeyError('Task id=%d not found' % id)
        elif t is tuple:
            ids = id
        elif t is list:
            ids = id.split(".")
        else:
            raise RegistryAccessError(
                'Expected a job id: int, (int,int), or "int.int"')

        if not len(ids) in [1, 2]:
            raise RegistryAccessError(
                'Too many ids in the access tuple, 2-tuple (job,subjob) only supported')

        try:
            ids = [int(id) for id in ids]
        except TypeError:
            raise RegistryAccessError(
                'Expeted a job id: int, (int,int), or "int.int"')
        except ValueError:
            raise RegistryAccessError(
                'Expected a job id: int, (int,int), or "int.int"')

        try:
            j = self.objects[ids[0]]
        except KeyError:
            raise RegistryKeyError('Task %d not found' % ids[0])

        if len(ids) > 1:
            try:
                return j.transforms[ids[1]]
            except IndexError:
                raise RegistryKeyError(
                    'Transform %s not found' % ('.'.join([str(id) for id in ids])))
        else:
            return j

    def remove(self, keep_going):
        self.do_collective_operation(keep_going, 'remove')

    def run(self, keep_going):
        self.do_collective_operation(keep_going, 'run')

    def pause(self, keep_going):
        self.do_collective_operation(keep_going, 'pause')


from Ganga.GPIDev.Lib.Registry.RegistrySliceProxy import RegistrySliceProxy, _wrap, _unwrap


class TaskRegistrySliceProxy(RegistrySliceProxy):

    """This object is an access list of tasks.

    The 'tasks' represents all existing tasks.

    A subset of tasks may be created by slicing (e.g. tasks[-10:] last ten tasks)
    or select (e.g. tasks.select(status='new') or tasks.select(10,20) tasks with
    ids between 10 and 20). A new access list is created as a result of
    slice/select. The new access list may be further restricted.

    This object allows to perform collective operations listed below such as
    run on all tasks in the current range. The keep_going=True
    (default) means that the operation will continue despite possible errors
    until all tasks are processed. The keep_going=False means that the
    operation will bail out with an Exception on a first encountered error.
    """

    def remove(self, keep_going=True):
        """ Remove all tasks."""
        return stripProxy(self).remove(keep_going=keep_going)

    def run(self, keep_going=True):
        """ Run all tasks."""
        return stripProxy(self).run(keep_going=keep_going)

    def pause(self, keep_going=True):
        """ Pause all tasks."""
        return stripProxy(self).pause(keep_going=keep_going)

    def copy(self, keep_going=True):
        """ Copy all tasks. """
        return JobRegistrySliceProxy(stripProxy(self).copy(keep_going=keep_going))

    def select(self, minid=None, maxid=None, **attrs):
        """ Select a subset of tasks. Examples:
        tasks.select(10): select tasks with ids higher or equal to 10;
        tasks.select(10,20) select tasks with ids in 10,20 range (inclusive);
        tasks.select(status='completed') select all tasks with status completed;
        tasks.select(name='some') select all tasks with some name;
        """
        unwrap_attrs = {}
        for a in attrs:
            unwrap_attrs[a] = _unwrap(attrs[a])
        return TaskRegistrySliceProxy(stripProxy(self).select(minid, maxid, **unwrap_attrs))

    def __call__(self, x):
        """ Access individual job. Examples:
        tasks(10) : get job with id 10 or raise exception if it does not exist.
        tasks((10,2)) : get transform number 2 of task 10 if exist or raise exception.
        tasks('10.2')) : same as above
        """
        return _wrap(stripProxy(self).__call__(x))

    def __getitem__(self, x):
        """ Get a job by positional index. Examples:
        tasks[-1] : get last job,
        tasks[0] : get first job,
        tasks[1] : get second job.
        """
        return _wrap(stripProxy(self).__getitem__(_unwrap(x)))

    # Information methods
    def table(self, id=None):
        """Prints a more detailed table of tasks and their transforms"""
        print("%s" % self.__str__(False, id))
        return

    def __str__(self, short=True, id=None):
        """Prints an overview over the currently running tasks"""
        if Ganga.GPIDev.Lib.Registry.RegistrySlice.config["tasks_show_help"]:
            self.help(short=True)
            Ganga.GPIDev.Lib.Registry.RegistrySlice.config.setUserValue("tasks_show_help", False)
            print("To show this help message again, type 'tasks.help()'.")
            print('')
            print(" The following is the output of " + markup("tasks.table()", fgcol("blue")))

        lenfstring = 0
        flist = []
        for thing in Ganga.GPIDev.Lib.Registry.RegistrySlice.config["tasks_columns"]:
            width = Ganga.GPIDev.Lib.Registry.RegistrySlice.config["tasks_columns_width"][thing]
            lenfstring += width
            flist.append("%" + str(width) + "s ")

        fstring = "|".join(flist)
        fstring += '\n'
        lenfstring += 27
        ds = fstring % ("#", "Type", "Name", "State", "Comment", "%4s: %4s/ %4s/ %4s/ %4s/ %4s/ %4s/ %4s" % ("Jobs", markup("done", overview_colours["completed"]), " " + markup("run", overview_colours["running"]), " " + markup("subd", overview_colours["submitted"]), " " + markup("attd", overview_colours["attempted"]), markup("fail", overview_colours["failed"]), markup("hold", overview_colours["hold"]), " " + markup("bad", overview_colours["bad"])), "Float")
        ds += "-" * lenfstring + "\n"


        from Ganga.GPIDev.Lib.Tasks import ITask

        if id is not None and type(id) is int:
            iterable_list = [self[id]]
        else:
            iterable_list = stripProxy(self).objects.values()

        for p in iterable_list:

            if isType(p, ITask):
                stat = "%4i: %4i/ %4i/  %4i/    --/ %4i/ %4i/ %4i" % (p.n_all(), p.n_status("completed"), p.n_status("running"), p.n_status("submitted"), p.n_status("failed"), p.n_status("hold"), p.n_status("bad"))
                ds += markup(fstring % (p.id, getName(p), p.name[0:Ganga.GPIDev.Lib.Registry.RegistrySlice.config['tasks_columns_width']['Name']], p.status, p.comment, stat, p.float), status_colours[p.status])
            else:
                stat = "%4i: %4i/ %4i/    --/  %4i/ %4i/ %4i/ %4i" % (p.n_all(), p.n_status("completed"), p.n_status("running"), p.n_status("attempted"), p.n_status("failed"), p.n_status("hold"), p.n_status("bad"))
                ds +=  markup(fstring % (p.id, getName(p), p.name[0:Ganga.GPIDev.Lib.Registry.RegistrySlice.config['tasks_columns_width']['Name']], p.status, p.comment, stat, p.float), status_colours[p.status])

            if short is True:
                continue

            for ti in range(0, len(p.transforms)):
                t = p.transforms[ti]

                if isType(p, ITask):
                    stat = "%4i: %4i/ %4i/ %4i/     --/ %4i/ %4i/ %4s" % (t.n_all(), t.n_status("completed"), t.n_status("running"), t.n_status("submitted"), t.n_status("failed"), t.n_status("hold"), t.n_status("bad"))
                    ds += "\n" + markup(fstring % ("%i.%i" % (p.id, ti), getName(t), t.name[0:Ganga.GPIDev.Lib.Registry.RegistrySlice.config['tasks_columns_width']['Name']], t.status, "", stat, ""), status_colours[t.status])
                else:
                    stat = "%4i: %4i/ %4i/     --/ %4i/ %4i/ %4i/ %4s" % (t.n_all(), t.n_status("completed"), t.n_status("running"), t.n_status("attempted"), t.n_status("failed"), t.n_status("hold"), t.n_status("bad"))
                    ds += "zn" + markup(fstring % ("%i.%i" % (p.id, ti), getName(t), t.name[0:Ganga.GPIDev.Lib.Registry.RegistrySlice.config['tasks_columns_width']['Name']], t.status, "", stat, ""), status_colours[t.status])

            ds += "-" * lenfstring + "\n"

        return ds + "\n"

    #_display = __str__
    def _display(self):
        return self.__str__(True)

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('tasks...')
            return
        p.text(self.__str__(True))


    def help(self, short=False):
        """Print a short introduction and 'cheat sheet' for the Ganga Tasks package"""
        print('')
        print(markup(" *** Ganga Tasks: Short Introduction and 'Cheat Sheet' ***", fgcol("blue")))
        print('')
        print(markup("Definitions: ", fgcol("red")) + "'Partition' - A unit of processing, for example processing a file or processing some events from a file")
        print("             'Transform' - A group of partitions that have a common Ganga Application and Backend.")
        print("             'Task'      - A group of one or more 'Transforms' that can have dependencies on each other")
        print('')
        print(markup("Possible status values for partitions:", fgcol("red")))
        print(' * "' + markup("ready", overview_colours["ready"]) + '"    - ready to be executed ')
        print(' * "' + markup("hold", overview_colours["hold"]) + '"     - dependencies not completed')
        print(' * "' + markup("running", overview_colours["running"]) + '"  - at least one job tries to process this partition')
        print(' * "' + markup("attempted", overview_colours["attempted"]) + '"- tasks tried to process this partition, but has not yet succeeded')
        print(' * "' + markup("failed", overview_colours["failed"]) + '"   - tasks failed to process this partition several times')
        print(' * "' + markup("bad", overview_colours["bad"]) + '"      - this partition is excluded from further processing and will not be used as input to subsequent transforms')
        print(' * "' + markup("completed", overview_colours["completed"]) + '" ')
        print('')

        def c(s):
            return markup(s, fgcol("blue"))
        print(markup("Important commands:", fgcol("red")))
        print(" Get a quick overview     : " + c("tasks") + "                  Get a detailed view    : " + c("tasks.table()"))
        print(" Access an existing task  : " + c("t = tasks(id)") + "          Remove a Task          : " + c("tasks(id).remove()"))
        print(" Create a new (MC) Task   : " + c("t = MCTask()") + "           Copy a Task            : " + c("nt = t.copy()"))
        print(" Show task configuration  : " + c("t.info()") + "               Show processing status : " + c("t.overview()"))
        print(" Set the float of a Task  : " + c("t.float = 100") + "          Set the name of a task : " + c("t.name = 'My Own Task v1'"))
        print(" Start processing         : " + c("t.run()") + "                Pause processing       : " + c("t.pause()"))
        print(" Access Transform id N    : " + c("tf = t.transforms[N]") + "   Pause processing of tf : " + c( "tf.pause()") + "  # This command is reverted by using t.run()")
        print(" Transform Application    : " + c("tf.application") + "         Transform Backend      : " + c("tf.backend"))
        print('')
        print(" Set parameter in all applications       : " + c("t.setParameter(my_software_version='1.42.0')"))
        print(" Set backend for all transforms          : " + c("t.setBackend(backend) , p.e. t.setBackend(LCG())"))
        print(" Limit on how often jobs are resubmitted : " + c("tf.run_limit = 4"))
        print(" Manually change the status of partitions: " + c("tf.setPartitionStatus(partition, 'status')"))
        print('')
        print(" For an ATLAS Monte Carlo Production Example and specific help type: " + c("MCTask?"))
        print(" For an ATLAS Analysis Example and help type: " + c("AnaTask?"))
        print('')

        if not True:
            #      if not short:
            print("ADVANCED COMMANDS:")
            print("Add Transform  at position N      : t.insertTransform(N, transform)")
            print("Remove Transform  at position N   : t.removeTransform(N)")
            print("Set Transform Application         : tf.application = TaskApp() #This Application must be a 'Task Version' of the usual application")
            print("   Adding Task Versions of Applications is easy, contact the developers to request an inclusion")




from abc import abstractmethod

from Ganga.Core import GangaAttributeError
from Ganga.GPIDev.Base.Objects import GangaObject
from Ganga.GPIDev.Schema import Schema, Version


class ICredentialRequirement(GangaObject):
    """
    Requirements objects specify what is needed by a particular object in Ganga

    All subclasses must specify ``info_class`` as well as all abstract methods
    """

    _schema = Schema(Version(1, 0))

    _category = 'CredentialRequirement'
    _name = 'ICredentialRequirement'
    _hidden = 1

    # This attribute refers to the ICredentialInfo subclass which can satisfy the requirements
    info_class = None  # type: Type[ICredentialInfo]

    def __init__(self, **kwargs):
        super(ICredentialRequirement, self).__init__()
        for key, value in kwargs.items():
            if key not in self._schema.allItemNames():
                raise GangaAttributeError('{class_name} does not have attribute called "{attr}"'.format(class_name=self.__class__.__name__, attr=key))
            setattr(self, key, value)

    @abstractmethod
    def encoded(self):
        """
        Return a string which encoded all the requirements.
        This string will be appended to the filename so it should be filesystem-friendly.
        """
        pass

    def __str__(self):
        """ Returns the repr as defined within __repr__ """
        return repr(self)

    def __repr__(self):
        """ This is a custom strinf repr of the class which can regenerate it on the IPython prompt. FIXME: This should rely on GangaObject. """
        items = ((name, getattr(self, name)) for name in self._schema.allItemNames())  # Name,value pairs for all schema items
        explicit_items = ((name, value) for name, value in items if value)  # Filter out any default values
        item_strings = ((name, repr(value)) for name, value in explicit_items)  # Stringify the values
        arg_strings = ('='.join(arg) for arg in item_strings)  # Make "name=value" strings
        arg_string = ', '.join(arg_strings)  # Make a full "a=1, b=2" string
        return '{name}({args})'.format(name=self.__class__.__name__, args=arg_string)

    def __hash__(self):
        """ Returns a hash of the name of the credential requirement and the location of the requirement on disk with additional encoding """
        return hash(self.__class__.__name__ + self.encoded())

    def __eq__(self, other):
        """ This compares 2 Credential Requirements based upon additional encoding of the credential properties """
        if other is None:
            return False
        return type(self) is type(other) and self.encoded() == other.encoded()




##############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: StagerDataset.py,v 1.6 2009-03-26 20:33:11 hclee Exp $
###############################################################################
# ATLAS input dataset plugin for using Athena/FileStager with local/batch jobs

import os
import os.path
import re
import socket
from tempfile import mkstemp

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import *
from Ganga.Utility.Config import getConfig, ConfigError
from Ganga.Utility.logging import getLogger
from GangaAtlas.Lib.ATLASDataset import DQ2Dataset
from GangaAtlas.Lib.Athena import dm_util
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import *
from Ganga.Lib.LCG import ElapsedTimeProfiler

from dq2.info import TiersOfATLAS
from dq2.clientapi.DQ2 import DQ2
from dq2.common.DQException import * 

from Ganga.Core.GangaThread import GangaThread
from Queue import Queue, Empty

#subprocess.py crashes if python 2.5 is used
#try to import subprocess from local python installation before an
#import from PYTHON_DIR is attempted some time later
try:
    import subprocess
except ImportError:
    pass

def find(dirs, pattern):
    """
    finds files within the given directories. Only the filename matching the 
    given pattern will be returned.
    """

    fpaths = []
    re_pat = re.compile(pattern)

    def __file_picker__(re_pat, dirname, names):
        for n in names:
            fpath = os.path.join(dirname,n)
            if re_pat.match(n):
                fpaths.append(fpath)
            else:
                logger.debug('ignore file: %s' % fpath)

    for dir in dirs:
        absdir = os.path.abspath( os.path.expandvars( os.path.expanduser(dir) ) )
        if not os.path.isdir(absdir):
            logger.warning('dir. not exist: %s' % absdir)
        else:
            os.path.walk(absdir, __file_picker__, arg=re_pat)

    return fpaths

def list_castor_files(dirs, pattern):
    '''Lists castor files in the given castor directory''' 
    fpaths = []

    re_pat = re.compile(pattern)

    re_dir = re.compile('^d[r|w|x|-]*$')

    def __is_dir__(fpath, fpath_mode):
        logger.debug('%s is a dir' % fpath)
        return re_dir.match(fpath_mode)

    def __file_picker__(re_pat, dirname):
        'looking for files iteratively into subdirectories' 

        my_fpaths = []

        fpath_modes = {} 

        cmd = 'rfdir %s | awk \'{print $1,$NF}\'' % dirname
        (rc, out, err) = execSyscmdSubprocess(cmd)

        if rc != 0:
            logger.error('error listing files: %s' % err)
        else:

            for l in map(lambda x:x.strip(), out.strip().split('\n')):
                (fpath_mode, name) = l.split()
                fpath_modes[name.strip()] = fpath_mode.strip()

            for n in fpath_modes.keys():
                fpath = os.path.join(dirname,n)
                if __is_dir__(fpath, fpath_modes[n]):
                    my_fpaths += __file_picker__(re_pat, fpath)
                else: 
                    if re_pat.match(n):
                        my_fpaths.append(fpath)
                    else:
                        logger.debug('ignore file: %s' % fpath)
        return my_fpaths

    ## top-level directories given by user
    for dir in dirs:
        fpaths += __file_picker__(re_pat, dir)

    return fpaths

def resolve_file_locations(dataset, sites=None, cloud=None, token='ATLASDATADISK', debug=False):
    '''
    Summarize the locations of files (in terms of sitename) of a dataset.
    If the sites argument is given, ignoring cloud and token arguments;
    otherwise using cloud and toke to retrieve sites from TiersOfATLAS.
    '''

    if not sites:
        logger.debug('resolving sites with token: %s' % token)
        sites = dm_util.get_srmv2_sites(cloud, token=token, debug=debug)

    logger.debug('checking replicas at sites: %s' % str(sites))

    replicas = {}
    # preparing the queue for querying lfn 
    wq = Queue(len(sites))
    for site in sites:
        wq.put(site)

    mylock = Lock()

    def worker(id):
        dq2 = DQ2()
        while not wq.empty():
            try:
                site = wq.get(block=True, timeout=1)
                replicaInfo = dq2.listFileReplicas(site, dataset)
                logger.debug('resolving dataset files at %s, no files: %d' % (site,len(replicaInfo[0]['content'])) )
                if replicaInfo:
                    mylock.acquire()
                    for guid in replicaInfo[0]['content']:
                        if guid not in replicas:
                            replicas[guid] = []
                        replicas[guid].append(site)
                    mylock.release()
            except Empty:
                pass
            except DQException as err:
                logger.warning(str(err))
                logger.warning('site %s excluded' % site)
                pass

    threads = []
    nthread = len(sites)
    if nthread > 10: nthread = 10

    for i in range(nthread):
        t = GangaThread(name='stager_ds_w_%d' % i, target=worker, kwargs={'id': i})
#        t.setDaemon(False)
        threads.append(t)

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    return replicas

## system command executor with subprocess
def execSyscmdSubprocess(cmd, wdir=os.getcwd()):
    '''executes system command vir subprocess module'''

    import subprocess

    exitcode = -999

    mystdout = ''
    mystderr = ''

    try:

        ## resetting essential env. variables
        my_env = os.environ

        if 'LD_LIBRARY_PATH_ORIG' in my_env:
            my_env['LD_LIBRARY_PATH'] = my_env['LD_LIBRARY_PATH_ORIG']

        if 'PATH_ORIG' in my_env:
            my_env['PATH'] = my_env['PATH_ORIG']

        if 'PYTHONPATH_ORIG' in my_env:
            my_env['PYTHONPATH'] = my_env['PYTHONPATH_ORIG']

        child = subprocess.Popen(cmd, cwd=wdir, env=my_env, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        (mystdout, mystderr) = child.communicate()

        exitcode = child.returncode

    finally:
        pass

    return (exitcode, mystdout, mystderr)

class StagerDatasetConfigError(ConfigError):
    '''An exception object for stager dataset configuration'''
    def __init__(self,what):
        ConfigError.__init__(self, what)
        
    def __str__(self):
        return "StagerDatasetConfigError: %s "%(self.what)

class StagerDataset(DQ2Dataset):
    '''Input dataset specification for using Athena/FileStager module to get input files for analysis.

    Usage:

      - work with DQ2 dataset given a DQ2 dataset 'user.RichardHawkings.0108175.topmix_Muon.AOD.v4':

        In [n]: j.inputdata = StagerDataset()
        In [n]: j.inputdata.type = 'DQ2'
        In [n]: j.inputdata.dataset += [ 'user.RichardHawkings.0108175.topmix_Muon.AOD.v4' ]

        Please note that 'DQ2' type works only when the current Ganga client is in the same domain as the SE
        associated with the DDM site defined by 'config.DQ2.DQ2_LOCAL_SITE_ID'.

      - work with CASTOR/DPM given a directory '/castor/cern.ch/grid/atlas/atlasgroupdisk/phys-top/dq2/user/RichardHawkings/user.RichardHawkings.0108175.topmix_Muon.AOD.v4' in CASTOR/DPM name space:

        In [n]: j.inputdata = StagerDataset()
        In [n]: j.inputdata.type = 'CASTOR'
        In [n]: j.inputdata.dataset += ['/castor/cern.ch/grid/atlas/atlasgroupdisk/phys-top/dq2/user/RichardHawkings/user.RichardHawkings.0108175.topmix_Muon.AOD.v4']

      - work with files in a given directory '/home/data/user.RichardHawkings.0108175.topmix_Muon.AOD.v4' on local disk:

        In [n]: j.inputdata = StagerDataset()
        In [n]: j.inputdata.type = 'LOCAL'
        In [n]: j.inputdata.dataset += ['/home/data/user.RichardHawkings.0108175.topmix_Muon.AOD.v4']

      For reading files from CASTOR/DPM, one can also ask FileStager to use Xrootd protocol to copy the file (thus then "xrdcp" is used instead of "rfcp"):

        In [n]: j.inputdata.xrootd_access = True
    '''

    _schema = Schema(Version(1,1), {
        'dataset': SimpleItem(defvalue = [], typelist=['str'], sequence=1, strict_sequence=0, doc='Dataset Name(s) or a root path in which the dataset files are located'),
        'guids'      : SimpleItem(defvalue = [], typelist=['str'], sequence=1, doc='GUID of Logical File Names'),
        'tagdataset' : SimpleItem(defvalue = [], typelist=['str'], sequence=1, strict_sequence=0, hidden=1, doc = 'Tag Dataset Name'),
        'use_aodesd_backnav' : SimpleItem(defvalue = False, doc = 'Use AOD to ESD Backnavigation',hidden=1),
        'names'      : SimpleItem(defvalue = [], typelist=['str'], sequence = 1, doc = 'Logical File Names to use for processing', hidden=0),
        'exclude_names'      : SimpleItem(defvalue = [], typelist=['str'], sequence = 1, copyable=0, doc = 'Logical File Names to exclude from processing', hidden=1),
        'number_of_files' : SimpleItem(defvalue = 0, doc = 'Number of files. ', copyable=0, hidden=1),
        'type'       : SimpleItem(defvalue = 'DQ2',  doc = 'Dataset type: DQ2 (refer to a DQ2 dataset), CASTOR/DPM (refer to a castor/dpm path) or LOCAL (refer to a local directory)', hidden=0),
        'xrootd_access': SimpleItem(defvalue = False,  doc = 'Sets to True for staging the dataset files via xrootd protocol.', hidden=0),
        'datatype'   : SimpleItem(defvalue = '', doc = 'Data type: DATA, MC or MuonCalibStream', hidden=0),
        'accessprotocol'       : SimpleItem(defvalue = '', doc = 'Accessprotocol to use on worker node, e.g. Xrootd', hidden=1),
        'match_ce_all' : SimpleItem(defvalue = False, doc = 'Match complete and incomplete sources of dataset to CE during job submission', hidden=1),
        'min_num_files' : SimpleItem(defvalue = 0, doc = 'Number of minimum files at incomplete dataset location', hidden=1),
        'complete_files_replicas' : SimpleItem(defvalue={},protected=1,copyable=0,transient=1,hidden=1,doc='The hidden cache of the info of the file replicas of the complete datasets'),
        'check_md5sum' : SimpleItem(defvalue = False, doc = 'Check md5sum of input files on storage elemenet - very time consuming !', hidden=1)
    })

    _category = 'datasets'
    _name = 'StagerDataset'
    _exportmethods = [ 'list_datasets', 'list_contents', 'get_surls', 'get_locations', 'get_file_locations', 'make_input_option_file', 'get_complete_files_replicas', 'make_FileStager_jobOptions' ]

    _GUIPrefs = [ { 'attribute' : 'dataset',     'widget' : 'String_List' }]

    def __init__(self):
        super( StagerDataset, self ).__init__()

    def __setattr__(self, attr, value):

        DQ2Dataset.__setattr__(self, attr, value)

        if attr == 'dataset':
            self.complete_files_replicas = {} 

    def __resolve_containers(self, containers, nthreads=10):
        '''resolving dataset containers'''

        datasets = {} 
        
        wq = Queue(len(containers))
        for ds in containers:
            wq.put(ds)

        mylock = Lock()
        def worker(id):
            dq2 = DQ2()
            while not wq.empty():
                try:
                    ds = wq.get(block=True, timeout=1)
                    logger.debug('worker id: %d on dataset container: %s' % (id, ds))
       
                    datasets[ds] = []
 
                    ds_tmp = dq2.listDatasetsInContainer(ds)

                    mylock.acquire()
                    datasets[ds] = ds_tmp
                    mylock.release()
                except DQException as err:
                    logger.warning(str(err))
                except Empty:
                    pass

        profiler = ElapsedTimeProfiler(logger=logger)
        profiler.start()
        threads = []
        for i in range(nthreads):
            t = GangaThread(name='stager_ds_w_%d' % i, target=worker, kwargs={'id': i})
#            t.setDaemon(False)
            threads.append(t)
        
        for t in threads:
            t.start()
        
        for t in threads:
            t.join()
        profiler.check('resolving %d containers' % len(containers))

        return datasets

    def __expand_datasets(self):

        logger.debug('resolving wildcard dataset patterns')

        # try to expand the dataset if a wildcard pattern is given
        expanded_datasets = []
        for dsn in self.dataset:
            if dsn.find('*') > -1:
                expanded_datasets += [ds[0] for ds in self.get_datasets(dsn)]
            else:
                expanded_datasets.append(dsn)

        # resolving container datasets
        containers = []
        datasets   = []
        for dsn in expanded_datasets:
            if dsn[-1] == '/':
                containers += [ dsn ]
            else:
                datasets += [ dsn ]

        if len(containers) > 0:
            logger.debug('resolving dataset containers')
            container_ds = self.__resolve_containers(containers)
            for c in container_ds.keys():
                datasets += container_ds[c]

        self.dataset = datasets

    def dataset_exists(self):
        return DQ2Dataset.dataset_exists(self) 

    def tagdataset_exists(self):
        logger.error('method not supported')
        return None

    def get_datasets(self, name, filter=True):
        '''Get datasets names, especially when the name is given as a wildcard pattern'''
        profiler = ElapsedTimeProfiler(logger=logger)
        profiler.start()
        datasets = listDatasets(name, filter)
        profiler.check('listing %d datasets/containers' % len(datasets))

        return datasets
    
    def get_contents(self,backnav=False, overlap=True):
        '''Helper function to access dataset content'''

        # expand datasets in case there is a wildcard 
        self.__expand_datasets()

        # always get all contents
        self.number_of_files = 0

        return DQ2Dataset.get_contents(self, backnav=backnav, overlap=overlap)

    def get_tag_contents(self):
        '''Helper function to access tag datset content'''

        logger.error('method not supported')
        return None 

    def get_locations(self, complete=0, backnav=False, overlap=True):
        '''helper function to access the dataset location'''

        # expand datasets in case there is a wildcard 
        self.__expand_datasets()

        logger.debug('getting dataset locations')

        return DQ2Dataset.get_locations(self,complete=complete, backnav=backnav, overlap=overlap) 

    def get_file_locations(self, complete=1, sites=None, cloud=None):
        '''Helper function to resolve replica locations of files of the given datasets
           
           NB: by default, it checks only sites with complete datasets   
        '''

        profiler = ElapsedTimeProfiler(logger=logger)

        # get the dataset locations 
        ds_sites = self.get_locations(complete=complete, overlap=False)

        logger.info('resolving dataset file locations')
        profiler.start()
        replicas = {}
         
        #if complete:
        #    pass
        #else:
        for ds in ds_sites.keys():
            logger.debug('dataset: %s' % ds)
            logger.debug('sites: %s' % repr(ds_sites[ds]))
            replicas.update(resolve_file_locations(ds, sites=ds_sites[ds]))
         
        profiler.check('%d datasets %d files' % ( len(ds_sites.keys()), len(replicas.keys()) ))

        return replicas

    def get_complete_files_replicas(self, nthread=10, diskOnly=True):
        '''Gets a comprehensive dataset information about the contents and the
           location of COMPLETE replicas'''

        if not self.complete_files_replicas:

            re_tapeSite = re.compile('.*TAPE$') 


            ds_info = {}
            self.__expand_datasets()
         
            wq = Queue(len(self.dataset))
            for ds in self.dataset:
                wq.put(ds)
         
            mylock = Lock()
            def worker(id):
         
                dq2 = DQ2()
                while not wq.empty():
                    try:
         
                        ds = wq.get(block=True, timeout=1)
                        logger.debug('worker id: %d on dataset: %s' % (id, ds))
         
                        # get contents (guids) of the complete dataset
                        contents = dq2.listFilesInDataset(ds)
         
                        # get locations of the complete dataset replicas
                        locations = dq2.listDatasetReplicas(ds,complete=1)
         
                        vuid = None
                        try:
                            vuid = locations.keys()[0]
                        except IndexError as err:
                            pass
         
                        mylock.acquire()
         
                        # updating ds_info hastable
                        if vuid:
                            ds_info[ds] = []
                            ds_sites = []
 
                            if diskOnly:
                                for site in locations[vuid][1]:
                                    if not re_tapeSite.match(site):
                                        ds_sites.append(site)
                            else:
                                ds_sites = locations[vuid][1]
 
                            ds_info[ds] += [ contents[0], ds_sites ]
                        else:
                            logger.warning('dataset not available: %s' % ds)
         
                        mylock.release()
         
                    except DQException as err:
                        logger.warning(str(err))
         
                    except Empty:
                        pass
         
            # prepare and run the query threads
            profiler = ElapsedTimeProfiler(logger=logger)
            profiler.start()
            threads = []
            for i in range(nthread):
                t = GangaThread(name='stager_ds_w_%d' % i, target=worker, kwargs={'id': i})
#                t.setDaemon(False)
                threads.append(t)
         
            for t in threads:
                t.start()
         
            for t in threads:
                t.join()

            self.complete_files_replicas = ds_info

            profiler.check( 'information collected: %d datasets' % ( len(self.complete_files_replicas.keys()) ) )
        else:
            logger.debug('using cached complete_files_replicas')
            pass
 
        return self.complete_files_replicas 

    def list_datasets(self,name,filter=True):
        '''List datasets names'''

        return DQ2Dataset.list_datasets(self, name=name, filter=filter)

    def list_contents(self,dataset=None):
        '''List dataset content'''

        self.__expand_datasets()

        return DQ2Dataset.list_contents(self, dataset=dataset)

    def list_locations(self,dataset=None,complete=0):
        '''List dataset locations'''

        # expand datasets in case there is a wildcard 
        self.__expand_datasets()

        return DQ2Dataset.list_locations(self, dataset=dataset, complete=complete)

    def list_locations_ce(self,dataset=None,complete=0):
        '''List the CE associated to the dataset location'''

        return DQ2Dataset.list_locations_ce(self, dataset=dataset, complete=complete)

    def list_locations_num_files(self,dataset=None,complete=-1,backnav=False):
        '''List the number of files replicated to the dataset locations'''

        return DQ2Dataset.list_locations_num_files(self, dataset=dataset, complete=complete, backnav=backnav) 

    def get_replica_listing(self,dataset=None,SURL=True,complete=0,backnav=False):
        '''Return list of guids/surl replicated dependent on dataset locations'''

        return DQ2Dataset.get_replica_listing(self, dataset=dataset, SURL=SURL, complete=complete, backnav=backnav)

    def list_locations_siteindex(self,dataset=None, timeout=15, days=2):
        return DQ2Dataset.list_locations_siteindex(self, dataset=dataset, timeout=timeout, days=days)

    def fill_guids(self):
        '''fill up the guids attribute of this StagerDataset object'''

        ds_info = self.get_complete_files_replicas()

        self.guids = []

        if ds_info:
            for ds in ds_info.keys():
                self.guids += ds_info[ds][0].keys()

    def get_surls(self, ddmSiteName=None, guidRefill=True):
        '''Gets a list of physical files located at certain site.
           Using DQ2_LOCAL_SITE_ID if site name is not specified explicitely.

           If the domain of the SE associated with the "ddmSiteName" is not matching the
           domain of the current Ganga client, an StagerDatasetConfigError exception is thrown.
           This is a protection to avoid local jobs to do lcg-cp across domains.'''

        pfns = {}
        if self.type in ['DQ2']:
            if ddmSiteName is None:
                ddmSiteName = config['DQ2_LOCAL_SITE_ID']
         
            srm_host = dm_util.get_srm_host(ddmSiteName)
            lfc_host = dm_util.get_lfc_host(ddmSiteName)
            cli_host = socket.getfqdn()
         
            if not srm_host or (not lfc_host):
                raise StagerDatasetConfigError('site information not found in ToA: %s' % ddmSiteName)

            # try to make a protection if srm_host and cli_host are not in the same domain
            # 1. determin the domainname patter by pasing "cli_host": the hostname of the Ganga client
            # 2. do a pattern matching on srm_host.
            # 3. if the matching on step 2 failed, srm_host and cli_host are not
            #    in the same domain (by protection) this should be failed
            cli_host_info = cli_host.split('.', 1)
            re_cli_domain = re.compile( '.*%s.*' % cli_host_info[-1] )

            if not re_cli_domain.match( srm_host ):
                raise StagerDatasetConfigError('SE (%s) not in the same domain as this client (%s)'% (srm_host, cli_host) )

            logger.debug('CLI Host: %s' % cli_host)
            logger.debug('SRM Host: %s' % srm_host)
            logger.debug('LFC Host: %s' % lfc_host)
         
            if guidRefill:
                self.fill_guids()
         
            # looks like self.guids is not a type of list
            # therfore a conversion is needed
            my_guids = []
            for guid in self.guids:
                my_guids.append(guid)
         
            pfns_all, cksum_all = dm_util.get_pfns(lfc_host, my_guids)
         
            for guid in self.guids:
                try:
                    for pfn in pfns_all[guid]:
                        if pfn.find(srm_host) > -1:
                            pfns[guid] = pfn
                except KeyError:
                    logger.warning('replica of guid:%s not found at %s' % (guid, ddmSiteName))

        elif self.type in ['DPM','CASTOR']:
            #fpaths = list_castor_files(dirs=self.dataset, pattern='.*\.root.*')
            fpaths = list_castor_files(dirs=self.dataset, pattern='.*')
            fpaths.sort()
            id = 0
            for fpath in fpaths:
                pfns[id] = fpath
                id += 1
            
        elif self.type in ['LOCAL']:
            #fpaths = find(dirs=self.dataset, pattern='.*\.root.*')
            fpaths = find(dirs=self.dataset, pattern='.*')
            fpaths.sort()
            id = 0
            for fpath in fpaths:
                pfns[id] = fpath
                id += 1

        return pfns

    def resolve_FileStager_Configurations(self, ddmSiteName=None):
        '''Resolves the PFNs into FileStager TURLs, e.g. srm://... -> gridcopy://srm://...'''

        turls        = []
        gridcopy     = False
        fs_cp_cmd    = ''
        fs_cp_args   = []
        fs_of_prefix = 'file:'

        if self.type in ['', 'DQ2']:

            if not self.names:

                guidRefill = False

                if not self.guids:
                    guidRefill = True
                
                pfns = self.get_surls(guidRefill=guidRefill, ddmSiteName=ddmSiteName)

                self.guids = pfns.keys()
                self.guids.sort()
                for guid in self.guids:
                    self.names.append( pfns[guid] )

            gridcopy  = True
            fs_cp_cmd = './fs-copy.py'
            for fpath in self.names:
                turls.append( '%s' % fpath )

        elif self.type in ['CASTOR','DPM']:
            
            if not self.names:
                self.names = self.get_surls().values()

            if self.xrootd_access:

                def __get_stage_host__(setype):
                    stage_host = ''
                    env_key = ''
                    try:
                        if setype in ['CASTOR']:
                            env_key = 'STAGE_HOST'
                        elif setype in ['DPM']:
                            env_key = 'DPM_HOST'
                        stage_host = os.environ[env_key]
                    except KeyError:
                        logger.warning('$%s not defined. xrootd_access disabled.' % env_key)
                    return stage_host

                ## try to resolve the xrootd host
                ##  - for CASTOR, it's the env. variable $STAGE_HOST
                ##  - for DPM, it's the env. variable $DPM_HOST (to be confirmed)
                stage_host = __get_stage_host__(self.type)

                if not stage_host:
                    self.xrootd_access = False

            if self.xrootd_access and stage_host:

                gridcopy     = True
                fs_cp_cmd    = 'xrdcp'
                fs_of_prefix = ''
                for fpath in self.names:
                    turls.append( 'root://%s/%s' % (stage_host, fpath) )
            else:

                gridcopy     = True
                fs_cp_cmd    = 'rfcp'
                fs_of_prefix = ''
                for fpath in self.names:
                    turls.append( '%s' % fpath )

        elif self.type in ['LOCAL']:

            gridcopy     = False
            fs_cp_cmd    = 'cp'
            fs_of_prefix = ''

            ## work through underlying directories to get '*.root*' files
            if not self.names:
                self.names = self.get_surls().values()
                
            for fpath in self.names:
                turls.append( '%s' % fpath )

        return (turls, gridcopy, fs_cp_cmd, fs_cp_args, fs_of_prefix)


    def make_FileStager_jobOptions(self, job=None, wdir=None, max_events=-1):
        '''makes FileStager job options and input collection job options'''

        (turls, gridcopy, cp_cmd, cp_args, of_prefix) = self.resolve_FileStager_Configurations()

        jo_name = 'FileStager_jobOption.py'
        ic_name = 'input.py'

        # determing the filepath of the input option file
        #  - if job object is given, store the job options in job's inputdir
        #  - else takes the wdir argument
        #  - else creates a temporary directory
        if job:
            wdir = job.inputdir
        else:
            if not wdir:
                wdir = tempfile.mkdtemp()

        jo_path  = os.path.join( wdir, jo_name )
        ic_path  = os.path.join( wdir, ic_name )

        if ( dm_util.make_FileStager_jobOption(turls, gridcopy=gridcopy, fs_cp_cmd=cp_cmd, fs_cp_args=cp_args, fs_of_prefix=of_prefix, maxEvent=max_events, ic_jo_path=ic_path, fs_jo_path=jo_path) ):

            if not os.path.exists( jo_path ):
                raise ApplicatonConfigurationError(None, 'job option files for FileStager not found: %s' % jo_path)

            if not os.path.exists( ic_path ):
                raise ApplicatonConfigurationError(None, 'job option files for FileStager not found: %s' % ic_path)

        else:
            raise ApplicatonConfigurationError(None, 'Unable to find/generate job option files for FileStager.')

        return (jo_path, ic_path)


logger = getLogger()
config = getConfig('DQ2')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ============================================================================
## @file
#  Application handler for BenderScript
#  @see https://twiki.cern.ch/twiki/bin/view/LHCb/BenderScriptTutorial
#
#  User specifies:
#  - script file(s) which contain BenderScript scripts,
#  - (optionally) configuration/Configurables files, to be used for 'importOptions'
#  - (optionally) set of command line arguments
# 
# At least one 'script' or 'import' file is required.
#
# The application executes the following line:
# @code   
# % bender {scripts} {arguments} --import {imports} --no-color --no-castor --import=data.py --batch
# @endcode     
# e.g.
# @code 
# % bender script1.py  --import import1.py -w -p5 --no-color --no-castor --import=data.py --batch
# @endcode
# where <code>data.py<code> a file with input data and xml catalog, automatically generated by Ganga/DIRAC
#
# @code
# my_script  = "~/cmtuser/tests/my_batch.py"
# my_app     = BenderScript (
#    version      = 'v28r3'       ,
#    scripts      =   my_script   ,
#    arguments    = [
#      '--tuplefile=Tuples.root' , ## the name of output file with n-tuples   (if needed)
#      '--histofile=Histos.root'   ## the name of output file with histograms (if needed)
#      ] 
#      ) 
# @endcode
#
# For full list of arguments try
# @code
# % bender -h 
# @endcode
# Note that <code>--no-castor</code>, <code>--no-color</code> and
# <code>--batch</code> will be appended automatically 
#
# @author Vladimir ROMANOVSKY  Vladimir.Romanovskiy@cern.ch
# @author Vanya BELYAEV        Ivan.Belyaev@itep.ru
# @fate 2016-03-16
#
# Version           $Revision:$
# Last Modification $Date:$
#                by $Author:$
# =============================================================================
"""The application handler for BenderScript

see https://twiki.cern.ch/twiki/bin/view/LHCb/BenderScriptTutorial

User specifies:
- script file(s) which contain BenderScript scripts,
- (optinally) configuration/Configurables files, to be used for 'importOptions'
- optionally) set of command line arguments

At least one 'script' or 'import' file is required.

The application executes the following line:

bender {scripts} {arguments} --import {imports} --no-color --no-castor --import=data.py --batch

e.g.

bender script1.py  --import import1.py -w -p5 --no-color --no-castor --import=data.py --batch

where data.py a file with input data and xml catalog, automatically generated by Ganga/DIRAC

The complete example:

my_script  = 'my_batch.py'
my_app     = BenderScript (
...  version      = 'v28r3'       ,
...  scripts      =   my_script   ,
...  arguments    = [
...      '--tuplefile=Tuples.root' , ## the name of output file with n-tuples 
...      '--histofile=Histos.root'   ## the name of output file with histograms
...  ]) 

For full list of arguments try
% bender -h 

Note that '--no-castor', '--no-color'and '--batch' will be appended automatically 

"""
# =============================================================================
__date__    = '2016-03-16'
__version__ = '$Revision:$'
__author__  = 'Vladimir ROMANOVSKY, Vanya BELYAEV'
# =============================================================================
import os
from   os.path                                 import split, join
from   Ganga.GPIDev.Schema.Schema              import FileItem, SimpleItem
from   Ganga.GPIDev.Lib.File                   import File
from   Ganga.Utility.util                      import unique
from   Ganga.Core                              import ApplicationConfigurationError
from   Ganga.GPIDev.Lib.File                   import ShareDir
from   Ganga.GPIDev.Lib.File.FileBuffer        import FileBuffer
from   GangaGaudi.Lib.Applications.GaudiBase   import GaudiBase
from   GangaGaudi.Lib.Applications.GaudiUtils  import fillPackedSandbox, gzipFile
from   Ganga.Utility.files                     import expandfilename, fullpath
from   Ganga.Utility.Config                    import getConfig
from   AppsBaseUtils                           import guess_version
from   Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

# Added for XML PostProcessing
from GangaLHCb.Lib.Applications import XMLPostProcessor

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

# =============================================================================
## the actual wrapper script to execute 
layout = """#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
from distutils.spawn import find_executable
bender_script = find_executable('bender')
import sys
# =============================================================================
## redefine arguments 
# =============================================================================
sys.argv += {scripts}        
sys.argv += {arguments}
sys.argv += [ '--import'  ] + {imports} 
sys.argv += [ '--no-color'  , '--no-castor', '--import=data.py' , '--batch']
sys.argv += [ '--command' ] + {command} 
##
# =============================================================================
## execute bender script 
# =============================================================================
execfile(bender_script)
# =============================================================================
# The END
# =============================================================================
"""
# =============================================================================
# @class BenderScript
# Application handler for BenderScript
#
# The user specifies:
#  - script file(s) which contain BenderScript script(s),
#  - (optionally) configuration/Configurables files, to be used for 'importOptions'
#  - (optionally) set of command-line arguments
# 
# At least one 'script' or 'import' file is required.
#
# The application executes the following line:
# @code    
# % bender {scripts} {arguments} --import {imports} --no-color --no-castor --import=data.py --batch
# @endcode    
# e.g.
# @code 
# % bender script1.py  --import import1.py -w -p5 --no-color --no-castor --import=data.py --batch
# @endcode
# where <code>data.py</code> a file with input data and xml catalog, automatically generated by Ganga/DIRAC
# @author Vladimir ROMANOVSKY  Vladimir.Romanovskiy@cern.ch
# @author Vanya BELYAEV        Ivan.Belyaev@itep.ru 
class BenderScript(GaudiBase):    
    """The application handler for BenderScript
    
    The user specifies:
    - script file(s) which contain BenderScript scripts,
    - configuration/Configurables files, to be used for 'importOptions'
    - set of command line arguments
    
    At least one 'script' or 'import' file is required.
    
    The application executes the following line:
    
    bender {scripts} {arguments} --import {imports} --no-color --no-castor --import=data.py --batch
    
    e.g.
    
    bender script1.py  --import import1.py -w -p5 --no-color --no-castor --import=data.py --batch
    
    where data.py a file with input data and xml catalog, automatically generated by Ganga/DIRAC
    
    """
     
    _name           = 'BenderScript'
    _category       = 'applications'
    _exportmethods  = GaudiBase._exportmethods[:]
    _exportmethods += ['prepare', 'unprepare']
    
    _schema = GaudiBase._schema.inherit_copy()
    
    _schema.datadict['package'] = SimpleItem(
        defvalue = None,
        typelist = ['str', 'type(None)'],
        doc      = """The package the application belongs to (e.g. 'Sim', 'Phys')
        """
        )
    _schema.datadict['masterpackage'] = SimpleItem (
        defvalue = None,
        typelist = [ 'str', 'type(None)' ],
        doc      = """The package where your top level requirements file is read from.
        Can be written either as a path 'Tutorial/Analysis/v6r0' or in traditional notation 
        'Analysis v6r0 Tutorial'
        """
        )
    
    _schema.datadict['setupProjectOptions'] = SimpleItem(
        defvalue = ''     ,
        typelist = [ 'str', 'type(None)'],
        doc      = """Extra options to be passed onto the SetupProject command
        used for configuring the environment. As an example 
        setting it to '--dev' will give access to the DEV area. 
        For full documentation of the available options see 
        https://twiki.cern.ch/twiki/bin/view/LHCb/SetupProject
        """
        )
    
    _schema.datadict['scripts'] = FileItem(
        preparable      = 1      ,
        sequence        = 1      ,
        strict_sequence = 0      ,
        defvalue        = []     ,
        doc             = """The names of the script files to execute.
        A copy will be made at submission time
        """
        )
    
    _schema.datadict['imports'] = FileItem (
        preparable      =  1     ,
        sequence        =  1     ,
        strict_sequence =  0     ,
        defvalue        = []     ,
        doc             = """The names of the files to be used for 'importOptions'.
        A copy will be made at submission time
        """
        )
    
    _schema.datadict['commands'] = SimpleItem(
        defvalue = []      ,
        typelist = ['str'] ,
        sequence =  1      ,
        doc      = """The commands to be executed,
        e.g. [ 'run(10)' , 'print ls()' , 'print dir()' ]
        """
        )
    
    _schema.datadict['arguments'] = SimpleItem(
        defvalue = []      ,
        typelist = ['str'] ,
        sequence =  1      ,
        doc      = """List of command-line arguments for bender script,
        e.g. ['-w','-p5'], etc.
        For python scripts and configuration/Configurable files for 'importOptions'
        it is much better to use the separate options 'scripts' and 'imports'
        Following arguments will be appended automatically:  --no-color, --no-castor and --batch
        """
        )
    
    _schema.version.major += 2
    _schema.version.minor += 0
    
    #def __init__(self):
    #    super(BenderScrip, self).__init__()

    def _get_default_version(self, gaudi_app):
        return guess_version(self, gaudi_app)

    def _auto__init__(self):
        if not self.appname : self.appname = 'Bender'
        self._init()

    def _getshell(self):

        import EnvironFunctions
        return EnvironFunctions._getshell(self)

    def prepare(self, force=False):

        super(BenderScript, self).prepare(force)
        self._check_inputs()

        
        share_dir = os.path.join (
            expandfilename ( getConfig('Configuration')['gangadir'] ) ,
            'shared'                            ,
            getConfig('Configuration')['user']  ,
            self.is_prepared.name               )
        
        input_sandbox_tar = os.path.join ( share_dir , 'inputsandbox',
                                           '_input_sandbox_%s.tar' % self.is_prepared.name ) 
        input_sandbox_tgz = os.path.join ( share_dir , 'inputsandbox',
                                           '_input_sandbox_%s.tgz' % self.is_prepared.name ) 
        
        fillPackedSandbox ( self.scripts + self.imports , input_sandbox_tar  ) 
        gzipFile          ( input_sandbox_tar , input_sandbox_tgz , True     )
        
        # add the newly created shared directory into the metadata system if
        # the app is associated with a persisted object
        self.checkPreparedHasParent(self)
        self.post_prepare()
        logger.debug("Finished Preparing Application in %s" % share_dir)

    def master_configure(self):
        return (None, StandardJobConfig())

    def configure(self, master_appconfig):
        
        ## strip leading and trailing blanks from arguments 
        self.arguments = [ a.strip() for a in self.arguments ]

        ## strip leading and trailing blanks from the command 
        self.commands  = [ a.strip() for a in self.commands  ]
        
        ## the script layout
        the_script    = layout.format (
            scripts   = [ os.path.join ( f.subdir , os.path.basename ( f.name ) ) for f in self.scripts ] , 
            imports   = [ os.path.join ( f.subdir , os.path.basename ( f.name ) ) for f in self.imports ] , 
            arguments = self.arguments  ,
            command   = self.commands    
            )

        print 'SCRIPT:\n', the_script
        
        # add summary.xml
        outputsandbox_temp  = XMLPostProcessor._XMLJobFiles()
        outputsandbox_temp += unique(self.getJobObject().outputsandbox)
        outputsandbox       = unique(outputsandbox_temp)
        
        input_files  = []
        input_files += [ FileBuffer('gaudipython-wrapper.py', the_script ) ]
        logger.debug("Returning StandardJobConfig")
        return (None, StandardJobConfig(inputbox=input_files,
                                        outputbox=outputsandbox))
    
    def _check_inputs(self):
        """Checks the validity of user's entries for BenderScript schema"""
        
        if not self.scripts and not self.imports : 
            raise ApplicationConfigurationError(None, "Application scripts/imports are not defined")
        
        if isinstance ( self.scripts , str ) : self.scripts = [ File ( self.scripts ) ]
        if isinstance ( self.imports , str ) : self.imports = [ File ( self.imports ) ]
        
        for f in self.scripts : f.name = fullpath ( f.name )
        for f in self.imports : f.name = fullpath ( f.name )

    
    def postprocess(self):
        XMLPostProcessor.postprocess(self, logger)

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
# Associate the correct run-time handlers to GaudiPython for various backends.

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers      import allHandlers
from GangaLHCb.Lib.RTHandlers.LHCbGaudiRunTimeHandler      import LHCbGaudiRunTimeHandler
from GangaLHCb.Lib.RTHandlers.LHCbGaudiDiracRunTimeHandler import LHCbGaudiDiracRunTimeHandler

for backend in ['LSF', 'Interactive', 'PBS', 'SGE', 'Local', 'Condor', 'Remote']:
    allHandlers.add('BenderScript', backend, LHCbGaudiRunTimeHandler)
allHandlers.add('BenderScript', 'Dirac', LHCbGaudiDiracRunTimeHandler)

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

# =============================================================================
if '__main__' == __name__ :

    print 80*'*'  
    print __doc__ 
    print ' Author  : %s ' %  __author__   
    print ' Version : %s ' %  __version__  
    print ' Date    : %s ' %  __date__     
    print 80*'*'  

# =============================================================================
# The END 
# =============================================================================



import os
#from multiprocessing     import cpu_count
from Ganga.Utility.Config import makeConfig, getConfig
from Ganga.Utility.logging import getLogger

from Ganga.Utility.Config.Config import _after_bootstrap

from Ganga.GPIDev.Credentials.CredentialStore import credential_store

logger = getLogger()

if not _after_bootstrap:
    configDirac = makeConfig('DIRAC', 'Parameters for DIRAC')
    config = getConfig('Configuration')

    # Set default values for the Dirac section.
    # configDirac.addOption('ShowDIRACstdout', False,
    #                      'Display DIRAC API stdout to the screen in Ganga?')

    configDirac.addOption('Timeout', 1000,
                      'Default timeout (seconds) for Dirac commands')

    configDirac.addOption('splitFilesChunks', 5000,
                      'when splitting datasets, pre split into chunks of this int')
    diracenv = ""
    if "GANGADIRACENVIRONMENT" in os.environ:
        diracenv = os.environ["GANGADIRACENVIRONMENT"]

    configDirac.addOption('DiracEnvFile', diracenv, 'DEPRECATED. Ganga environment file for DIRAC environment (do not change unless you are sure you know what you are doing).')

    configDirac.addOption('DiracEnvSource', None, 'File to be sourced to provide the DIRAC environment. E.g. /cvmfs/ganga.cern.ch/dirac_ui/bashrc')
    configDirac.addOption('DiracEnvJSON', None, 'A JSON file containing the environment for DIRAC. Overrides DiracEnvSource')

    configDirac.addOption('DiracCommandFiles', [os.path.join(os.path.dirname(__file__), 'Lib/Server/DiracDefinition.py'),
                                                os.path.join(os.path.dirname(__file__), 'Lib/Server/DiracCommands.py')],
                      'The file containing the python commands that the local DIRAC server can execute. The default DiracCommands.py is added automatically')

    configDirac.addOption('noInputDataBannedSites', [],
                      'List of sites to ban when a user job has no input data (this is meant to reduce the load on these sites)')

    configDirac.addOption('MaxDiracBulkJobs', 500, 'The Maximum allowed number of bulk submitted jobs before Ganga intervenes')

    configDirac.addOption('failed_sandbox_download', True, 'Automatically download sandbox for failed jobs?')

    configDirac.addOption('load_default_Dirac_backend', True, 'Whether or not to load the default dirac backend. This allows packages to load a modified version if necessary')

    ## TODO this would be nice to move from here once the credentials branch is up and running
    configDirac.addOption('userVO', '', 'The name of the VO that the user belongs to')

    configDirac.addOption('DiracLFNBase', '', "Base dir prepended to create LFN name from DiracFile('name'). If this is unset then it will default to /[userVO]/user/[first letter of user name]/[user name]")

    configDirac.addOption('ReplicateOutputData', False,
                      'Determines whether outputdata stored on Dirac is replicated')

    configDirac.addOption('allDiracSE', [], 'SE/Space-Tokens allowed for replication, writing files etc.')

    configDirac.addOption('DiracFileAutoGet', True, 'Should the DiracFile object automatically poll the Dirac backend for missing information on an lfn?')

    configDirac.addOption('OfflineSplitterFraction', 0.75, 'If subset is above OfflineSplitterFraction*filesPerJob then keep the subset')
    configDirac.addOption('OfflineSplitterMaxCommonSites', 3, 'Maximum number of storage sites all LFN should share in the same dataset. This is reduced to 1 as the splitter gets more desperate to group the data.')
    configDirac.addOption('OfflineSplitterUniqueSE', True, 'Should the Sites chosen be accessing different Storage Elements.')
    configDirac.addOption('OfflineSplitterLimit', 50,
                      'Number of iterations of selecting random Sites that are performed before the spliter reduces the OfflineSplitter fraction by raising it by 1 power and reduces OfflineSplitterMaxCommonSites by 1. Smaller number makes the splitter accept many smaller subsets higher means keeping more subsets but takes much more CPU to match files accordingly.')

    configDirac.addOption('RequireDefaultSE', True, 'Do we require the user to configure a defaultSE in some way?')

    configDirac.addOption('statusmapping', {'Checking': 'submitted',
                                            'Completed': 'running',
                                            'Deleted': 'failed',
                                            'Done': 'completed',
                                            'Failed': 'failed',
                                            'Killed': 'failed',
                                            'Matched': 'submitted',
                                            'Received': 'submitted',
                                            'Running': 'running',
                                            'Staging': 'submitted',
                                            'Stalled': 'running',
                                            'Waiting': 'submitted'}, "Mapping between Dirac Job Major Status and Ganga Job Status")

    configDirac.addOption('finalised_statuses',
                                                {'Done': 'completed',
                                                 'Failed': 'failed',
                                                 'Killed': 'failed',
                                                 'Deleted': 'failed',
                                                 'Unknown: No status for Job': 'failed'},
                                                "Mapping of Dirac to Ganga Job statuses used to construct a queue to finalize a given job, i.e. final statues in 'statusmapping'")

    configDirac.addOption('serializeBackend', False, 'Developer option to serialize Dirac code for profiling/debugging')


def standardSetup():

    import PACKAGE
    PACKAGE.standardSetup()



def loadPlugins(config=None):
    logger.debug("Loading Backends")
    import Lib.Backends
    logger.debug("Loading RTHandlers")
    import Lib.RTHandlers
    logger.debug("Loading Files")
    import Lib.Files

def postBootstrapHook():
    dirac_conf = getConfig('DIRAC')
    if not dirac_conf['DiracEnvSource'] and not dirac_conf['DiracEnvJSON']:
        return
    from GangaDirac.Lib.Credentials.DiracProxy import DiracProxy
    try:
        credential_store[DiracProxy()]
    except KeyError:
        pass




from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.Backends.DiracBase import DiracBase
from GangaDirac.Lib.Backends.DiracUtils import result_ok
from Ganga.GPIDev.Schema import Schema, Version, ComponentItem
from Ganga.Core import BackendError
from Ganga.Core import GangaException
from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
from GangaDirac.Lib.Utilities.DiracUtilities import execute
import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()
from Ganga.GPIDev.Credentials              import require_credential


class Dirac(DiracBase):
    _schema = DiracBase._schema.inherit_copy()

    _schema.version.major += 0
    _schema.version.minor += 0
    _exportmethods = DiracBase._exportmethods[:]
    _exportmethods += ['checkSites', 'checkTier1s']
    _packed_input_sandbox = True
    _category = "backends"
    _name = 'Dirac'
    __doc__ = DiracBase.__doc__

    def _addition_sandbox_content(self, subjobconfig):
        input_sandbox = []
        j = self.getJobObject()
        for f in j.inputfiles.get(DiracFile):
            if f.lfn == '':
                raise GangaException(
                    'Can not add the lfn of of the DiracFile with name pattern: %s as the lfn property has not been set.' % f.namePattern)
            else:
                input_sandbox.append('LFN:' + f.lfn)
        return input_sandbox

    def _setup_subjob_dataset(self, dataset):
        return LHCbDataset(files=[DiracFile(lfn=f) for f in dataset])

    @require_credential
    def checkSites(self):
        cmd = 'checkSites()'
        result = execute(cmd, cred_req=self.credential_requirements)
        if not result_ok(result):
            logger.warning('Could not obtain site info: %s' % str(result))
            return
        return result.get('Value', {})

    @require_credential
    def checkTier1s(self):
        cmd = 'checkTier1s()'
        result = execute(cmd, cred_req=self.credential_requirements)
        if not result_ok(result):
            logger.warning('Could not obtain Tier-1 info: %s' % str(result))
            return
        return result.get('Value', {})

    def getOutputDataLFNs(self):
        """Get a list of outputdata that has been uploaded by Dirac. Excludes
        the outputsandbox if it is there."""
        lfns = super(Dirac, self).getOutputDataLFNs()
        ds = LHCbDataset()
        for f in lfns:
            ds.files.append(DiracFile(lfn=f))
        return GPIProxyObjectFactory(ds)

    def getOutputData(self, outputDir=None, names=None, force=False):
        """Retrieve data stored on SE to outputDir (default=job output workspace).
        If names=None, then all outputdata is downloaded otherwise names should
        be a list of files to download. If force is True then download performed
        even if data already exists."""
        downloaded_files = super(Dirac, self).getOutputData(outputDir, names, force)
        ds = LHCbDataset()
        for f in downloaded_files:
            ds.files.append(DiracFile(lfn=f))
        return GPIProxyObjectFactory(ds)



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ColourText.py,v 1.1 2008-07-17 16:41:00 moscicki Exp $
##########################################################################

# File: ColourText.py
# Author: K. Harrison
# Created: 050824
# Last modified: 050824

"""Module containing classes to help with creating colour text.

   Classes for text markup:

      ANSIMarkup - apply ANSI codes to the text
      NoMarkup - ingore colour codes and leave the text unchanged

   ANSI code classes defined are:

      Foreground - create object carrying ANSI codes for
                   changing text foreground colour;

      Background - create object carrying ANSI codes for
                   changing text background colour;

      Effects    - create object carrying ANSI codes for
                   changing text effects.

   Example usage is as follows:

      fg = Foreground()
      bg = Background()
      fx = Effects()

      if coloring_enabled:
       # text will be blue by default and colour codes are applied
       m = ANSIMarkup(fg.blue)
      else:
       # colour codes are ignored
       m = NoMarkup()

       # Text will be coloured in red if coloring_enabled,
       # otherwise text will be unchanged.
       print(m('Text in red',code=fg.red))

   It is a better practice to use markup objects to apply colours
   because it is easier to enable/disable the markup if desired.

   However inserting the codes directly also works:

   # Print text in specified colour.
   print(fg.some_colour + 'Text in some_colour' + fx.normal)

   # Print text with colour of background changed as specified.
   print(bg.some_colour + 'Text with background in some_colour' + fx.normal)

   # Print text with specified effect applied.
   print(fx.some_effect + 'Text with some_effect applied' + fx.normal)

   Note that each ANSI code overrides the previous one, and their effect
   isn't cumulative.  Terminating a string with fx.normal in the above
   examples ensures that subsequent text is output using the terminal
   defaults.

   For details of the colours and effects available, see help for
   individual classes.
"""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "24 August 2005"
__version__ = "1.0"


class ANSIMarkup(object):

    """ Apply ANSI colouring codes.
    """

    def __init__(self, default_code=None):
        if default_code is None:
            default_code = Effects().normal
        self.default_code = default_code

    def __call__(self, text, code=None):
        if code is None:
            code = self.default_code
        return code + text + self.default_code


class NoMarkup(object):

    """ Leave text unchanged.
    """

    def __init__(self, default_code=None):
        pass

    def __call__(self, text, code=None):
        return text


class Background(object):

    """Class for creating objects carrying ANSI codes for changing
       text background colour.  The defined colours are:

       black, blue, cyan, green, orange, magenta, red, white.

       In all cases, the text foregreound colour is the terminal
       default (usually black or white)."""

    def __init__(self):
        """Set ANSI codes for defined colours"""

        _base = '\033[%sm'
        self.black = _base % "0;40"
        self.red = _base % "0;41"
        self.green = _base % "0;42"
        self.orange = _base % "0;43"
        self.blue = _base % "0;44"
        self.magenta = _base % "0;45"
        self.cyan = _base % "0;46"
        self.white = _base % "0;47"


class Effects(object):

    """Class for creating objects carrying ANSI codes for text
       effects.  The defined effects are:

       normal,

       bold, reverse, underline,

       nobold, noreverse, nounderline.

       All effects imply terminal defaults for the colours."""

    def __init__(self):
        """Set ANSI codes for defined effects"""
        _base = '\033[%sm'
        self.normal = _base % "0;0"
        self.bold = _base % "0;1"
        self.underline = _base % "0;4"
        self.reverse = _base % "0;7"
        self.nobold = _base % "0;21"
        self.nounderline = _base % "0;24"
        self.noreverse = _base % "0;27"


class Foreground(object):

    """Class for creating objects carrying ANSI codes for changing
       text foreground colour.  The defined colours are:

       black, blue, cyan, green, orange, magenta, red, white,

       boldblue, boldcyan, boldgreen, boldgreen, boldgrey,
       boldmagenta, boldred, boldwhite, boldyellow.

       For good visibility, the bold colours are better."""

    def __init__(self):
        """Set ANSI codes for defined colours"""

        _base = '\033[%sm'
        self.normal = _base % "0"
        self.black = _base % "0;30"
        self.red = _base % "0;31"
        self.green = _base % "0;32"
        self.orange = _base % "0;33"
        self.blue = _base % "0;34"
        self.magenta = _base % "0;35"
        self.cyan = _base % "0;36"
        self.white = _base % "0;37"
        self.boldgrey = _base % "1;30"
        self.boldred = _base % "1;31"
        self.boldgreen = _base % "1;32"
        self.boldyellow = _base % "1;33"
        self.boldblue = _base % "1;34"
        self.boldmagenta = _base % "1;35"
        self.boldcyan = _base % "1;36"
        self.boldwhite = _base % "1;37"


colour_objects = {'fg': Foreground(), 'bg': Background(), 'fx': Effects()}

# Some additional helper functions from Tasks package
def getColour(name):
    """ Get a colour code from the symbolic name: fg = Foreground(), bg = Background(), fx = Effects()
    The name examples fg.red, fx.normal, bg.white
    Raise ValueError if name undefined or malformed.
    """
    x, y = name.split('.')
    return getattr(colour_objects[x], y)

cols = {"black": (0, 0), "red": (0, 1), "green": (0, 2), "orange": (0, 3), "blue": (0, 4), "magenta": (0, 5), "cyan": (0, 6), "lgray": (0, 7),
        "dgray": (6, 0), "lred": (6, 1), "lgreen": (6, 2), "yellow": (6, 3), "lblue": (6, 4), "pink": (6, 5), "lcyan": (6, 6), "white": (6, 7)}


def col(f, b):
    return '\033[%i%i;%i%im' % (4 + cols[b][0], cols[b][1], 3 + cols[f][0], cols[f][1])


def fgcol(f):
    return '\033[%i%im' % (3 + cols[f][0], cols[f][1])
# Status and overview colours
status_colours = {
    'new': "",
    'running': fgcol("green"),
    'completed': fgcol("blue"),
    'pause': fgcol("cyan"),
    'running/pause': fgcol("cyan"),
}
overview_colours = {
    'ignored': "",
    'hold': col("black", "lgray"),
    'ready': col("lgreen", "lgray"),
    'running': col("black", "green"),
    'completed': col("white", "blue"),
    'attempted': col("black", "yellow"),
    'failed': col("black", "lred"),
    'bad': col("red", "lcyan"),
    'unknown': col("white", "magenta"),
    'submitted': col("black", "lgreen"),
}


################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PACKAGE.py,v 1.2 2008-08-21 12:35:21 hclee Exp $
################################################################################

""" Refer to Ganga/PACKAGE.py for details on the purpose of this module.
"""
external_packages = {
   #'matplotlib' : {'version' : '0.99.0', 'PYTHONPATH':'lib/python2.5/site-packages'},
   #'numpy'      : {'version' : '1.3.0', 'PYTHONPATH':'lib/python2.5/site-packages', 'PATH' : 'bin'},
   #'pyqt'       : {'version' : '3.18.1_python2.5', 'PYTHONPATH':'lib/python2.5/site-packages', 'LD_LIBRARY_PATH' :'lib'}
   }

from Ganga.Utility.Setup import PackageSetup

setup = PackageSetup(external_packages)

def standardSetup(setup=setup):
    for p in setup.packages:
        setup.prependPath(p,'PYTHONPATH')
        setup.prependPath(p,'LD_LIBRARY_PATH')
        setup.prependPath(p,'PATH')



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
###############################################################################
# AthenaMC Local Runtime Handler
#

import os, string, commands, re, pwd

import imp,xml.dom.minidom,urllib,random

from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import *
from Ganga.Core import FileWorkspace
from Ganga.Core.exceptions import ApplicationConfigurationError

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from Ganga.Utility.Config import getConfig, ConfigError
from Ganga.Utility.logging import getLogger

from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

from GangaAtlas.Lib.ATLASDataset import DQ2Dataset
from GangaAtlas.Lib.ATLASDataset import DQ2OutputDataset
from GangaAtlas.Lib.AthenaMC.AthenaMCDatasets import extractFileNumber, matchFile


from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

# the config file may have a section
# aboout monitoring

mc = getConfig('MonitoringServices')

# None by default

#mc.addOption('AthenaMC', None, 'FIXME')
mc.addOption('AthenaMC/Local', None, 'FIXME')

_defaultSite='CERN-PROD_SCRATCHDISK'

class AthenaMCLocalRTHandler(IRuntimeHandler):
    """Athena MC LCG Runtime Handler"""

    def sortSites(self,insites,outsite):
        inlist=insites.split(" ")
        imax=outsite.find("_")
        outsite=outsite[:imax]
        newlist=[]
        for site in inlist:
            if site.find(outsite)>-1:
                newlist.insert(0,site)
            else:
                newlist.append(site)
        return string.join(newlist," ")
            
    def master_prepare(self,app,appmasterconfig):
        if app.siteroot: 
            os.environ["SITEROOT"]=app.siteroot
        if app.cmtsite:    
            os.environ["CMTSITE"]=app.cmtsite

        job = app._getParent()
        if app.dryrun:
            os.environ["SITEROOT"]  = "NONE"
            os.environ["CMTSITE"]  = "NONE"
        try:
            assert "SITEROOT" in os.environ
        except:
            raise ApplicationConfigurationError(None," ATLAS environment not defined")
                
        try:
            assert "CMTSITE" in os.environ and os.environ["CMTSITE"]!=""
        except:
            raise ApplicationConfigurationError(None,"cmt not setup properly. Please check your ATLAS setup or run on the grid")
            
        if "AtlasVersion" in os.environ:
            logger.debug("Checking AtlasVersion: %s and selected atlas release %s" % (os.environ["AtlasVersion"],app.atlas_rel))
            try:
                assert app.atlas_release==os.environ["AtlasVersion"]
            except:
                logger.error("Mismatching atlas release. Local setup is %s, resetting requested atlas release to local value." % os.environ["AtlasVersion"])
                app.atlas_release=os.environ["AtlasVersion"]
        elif "ATLAS_RELEASE" in os.environ:
            logger.debug("Checking ATLAS_RELEASE: %s and selected atlas release %s" % (os.environ["ATLAS_RELEASE"],app.atlas_rel))
            try:
                assert app.atlas_rel==os.environ["ATLAS_RELASE"]
            except:
                logger.error("Mismatching atlas release. Local setup is %s, resetting requested atlas release to local value." % os.environ["ATLAS_RELEASE"])
                app.atlas_rel=os.environ["ATLAS_RELEASE"]
        else:
            logger.warning("Could not compare requested release and local setup. Hope you are doing something sensible...")

                
        if job.backend._name=="LSF":
            try:
                assert "CMTSITE" in os.environ and os.environ["CMTSITE"]=="CERN"
            except:
                raise ApplicationConfigurationError(None,"Error, CERN ATLAS AFS environment not defined. Needed by LSF backend")

       
        environment={'T_LCG_GFAL_INFOSYS' :'atlas-bdii.cern.ch:2170'}
        for tag in os.environ:
            if os.environ[tag]!="":
                environment[tag]=os.environ[tag]
        
        trfopts=app.transflags
        # need to parse them to be able to pass them in an environment variable
        trfopts=trfopts.replace(" ","/W")
        trfopts=trfopts.replace("-","/F")
        
        trflags=trfopts
        if app.mode =="evgen":
            trflags="/Ft"
            if app.verbosity:
                trflags+="/W/Fl/W%s" % app.verbosity
        
        if trflags:
            environment["TRFLAGS"]=trflags

        # setting output site from input data if any.
        outsite,backup,outputlocation,backuplocation="","","",""
        logger.info("checking sites from input data: %s" % str(app.sites))

        # must distinguish running site (backend.requirements.sites) and output storage site (app.se_name)
        
        # matching with user's wishes (app.se_name or backend.requirements.sites)

        # select sites which are matching user's wishes, if any.
        selectedSites=app.sites
        print app.sites
        if job.backend._name!="Local":
            if len(selectedSites)>0:
                [outlfc,outsite,outputlocation]=job.outputdata.getDQ2Locations(selectedSites[0])
            if len(selectedSites)>1:
                [outlfc2,backup,backuplocation]=job.outputdata.getDQ2Locations(selectedSites[1])

        # app.se_name set: users wishes to get the output data written to another site than the one hosting the input.
        # One needs to ensure that this location is at least in the same cloud as the targetted processing site. This is done by insuring that the lfcs are the same.
        userSEs=[]
        outse=""
        if job.application.se_name and job.application.se_name != "none" and job.backend._name!="Local":
            userSEs=job.application.se_name.split(" ")
            # loop through userSEs until up to 2 valid sites are found...
            outse=""
            for SE in userSEs:
                [lfc,se,location]=job.outputdata.getDQ2Locations(SE)
                if lfc==outlfc:
                    if not outse:
                        outse=se # important to use outse and not outsite here, as outsite is used for selection of processing site.
                        # userSEs overrides outlfc,outputlocation, but not outsite as outsite is unfortunately used for choice of the processing site.
                        outputlocation=location
                    else:
                        outlfc2=lfc
                        backup=se
                        backuplocation=location
                        break
        
        # finally: if no backup location is defined at this point, enforce CERN-PROD_SCRATCHDISK as backup location
        if outsite=="":
            [outlfc,outsite,outputlocation]=job.outputdata.getDQ2Locations(_defaultSite)
        if backup=="":
             [outlfc2,backup,backuplocation]=job.outputdata.getDQ2Locations(_defaultSite)
        
        logger.info("Final selection of output sites: %s , backup: %s" % (outsite,backup))



        # srmv2 sites special treatment: the space token has been prefixed to the outputlocation and must be removed now:
        imin=string.find(outputlocation,"token:")
        imax=string.find(outputlocation,"srm:")
        spacetoken=""
        if imin>-1 and imax>-1:
            spacetoken=outputlocation[imin+6:imax-1]
            outputlocation=outputlocation[imax:]
        # same treatment for backup location if any
        imin=string.find(backuplocation,"token:")
        imax=string.find(backuplocation,"srm:")
        bst=""
        if imin>-1 and imax>-1:
            bst=backuplocation[imin+6:imax-1]
            backuplocation=backuplocation[imax:]

        environment["OUTLFC"]=outlfc
        environment["OUTSITE"]=outsite
        if outse:
            environment["OUTSITE"]=outse # user's choice for output storage location overriding AthenaMC's.
            
        environment["OUTPUT_LOCATION"]=outputlocation
        if spacetoken:
            environment["SPACETOKEN"]=spacetoken
        if backup:
            environment["OUTLFC2"]=outlfc2
            environment["OUTSITE2"]=backup
            environment["OUTPUT_LOCATION2"]=backuplocation

        environment["PROD_RELEASE"]=app.prod_release

        # setting environment["BACKEND"]
        # Local, Condor become "batch". LSF becomes "batch" unless the inputdata is on castor (in this case, it becomes "castor")
        environment["BACKEND"]=job.backend._name
        if job.backend._name=="LSF" and len(app.turls.values())>0:
            turl=app.turls.values()[0]
##            if string.find(turl,"castor")>-1:
##                environment["BACKEND"]="castor"
##            else:
##                environment["BACKEND"]="batch"
        if job.backend._name in ["Local","Condor","PBS"]:
            environment["SITEROOT"]=os.environ["SITEROOT"]
            environment["CMTSITE"]=os.environ["CMTSITE"]
            if job.backend._name in ["Condor","PBS"]:
                environment["BACKEND"]="batch"

#       finalise environment


        # preparing input sandbox, output sandbox , environment vars and job requirements
        
        inputbox = [ 
            File(os.path.join(os.path.dirname(__file__),'setup-release.sh')),
            File(os.path.join(os.path.dirname(__file__),'stage-in.sh')),
            File(os.path.join(os.path.dirname(__file__),'stage-out.sh')),
            File(os.path.join(os.path.dirname(__file__),'adler32.py'))
        ]

        if os.path.exists(app.transform_archive):
            # must add a condition on size.
            inputbox += [ File(app.transform_archive) ]
        elif app.transform_archive:
            # tarball in local or remote web area.
            if string.find(app.transform_archive,"http")>=0:
                environment['TRANSFORM_ARCHIVE'] = "%s" % (app.transform_archive)
            else:
                myfile=os.path.basename(app.transform_archive)
                myfile="http://cern.ch/atlas-computing/links/kitsDirectory/Production/kits/"+myfile
                environment['TRANSFORM_ARCHIVE'] = "%s" % (myfile)
                
        
        if app.evgen_job_option and os.path.exists(app.evgen_job_option):
            # locally modified job option file to add to the input sand box
            inputbox += [ File(app.evgen_job_option) ]
            # need to strip the path away.
            self.evgen_job_option = app.evgen_job_option.split("/")[-1]
            environment['CUSTOM_JOB_OPTION'] = "%s" % (self.evgen_job_option)
        elif app.evgen_job_option:
            self.evgen_job_option = app.evgen_job_option
            
        if (job.inputsandbox):
            for file in job.inputsandbox:
                inputbox += [ file ]


        outputbox = [ ]
        outputGUIDs='output_guids'
        outputLOCATION='output_location'
        outputbox.append( outputGUIDs )
        outputbox.append( outputLOCATION )
        outputbox.append( 'output_data' )
        if (job.outputsandbox):
            for file in job.outputsandbox:
                outputbox += [ file ]

        # switch JobTransforms/AtlasProduction package.
        self.isJT=string.find(app.transform_archive,"JobTransform")
        if self.isJT>-1 and app.mode=="evgen":
            environment['T_CONTEXT'] = str(self.number_events_job) # needed to avoid prodsys failure mechanism based on a hardcoded minimum number of event of 5000 per job


        # backend specifics:
        if job.backend._name=="SGE" and job.backend.extraopts=="":
            job.backend.extraopts="-l h_vmem=5G -l s_vmem=5G -l h_cpu=1:00:00 -l h_fsize=10G" # minimum set up for Atlfast II
            
        logger.debug("master job submit?")
        
        return StandardJobConfig("",inputbox,[],outputbox,environment)



        
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        """Prepare the job"""

        inputbox=[ ]

        #       prepare environment
        environment={}
        environment=jobmasterconfig.env.copy()
        environment["INPUTDATASETS"]=""
        environment["INPUTSITES"]=""
        environment["INPUTFILES"]=""

        alllfns=app.inputfiles+app.cavernfiles+app.mbfiles+app.dbfiles
        infilenr=0
        for infile in alllfns:
            environment["INPUTFILES"]+="lfn[%d]='%s';" %(infilenr,infile)
            environment["INPUTDATASETS"]+="dset[%d]='%s';"%(infilenr,app.dsetmap[infile])
            insites=app.sitemap[infile]
            # compare with environment["OUTSITE"] and reorder if needed.
            newinsites=self.sortSites(insites,environment["OUTSITE"])
            environment["INPUTSITES"]+="site[%d]='%s';"%(infilenr,newinsites)
            infilenr += 1


                
        logger.debug("%s %s %s" % (str(environment["INPUTDATASETS"]),str(environment["INPUTSITES"]),str(environment["INPUTFILES"])))

        job = app._getParent() # Returns job or subjob object
        # if datasetType is DQ2, then one needs to ensure that DQ2 environment is properly set.
        if job.inputdata:
            if job.inputdata.datasetType=="DQ2":
                environment["BACKEND"]="LCG"
                try:
                    assert "DQ2_LOCAL_SITE_ID" in os.environ
                except:
                    logger.error("Error in DQ2 configuration. Please leave ganga, then rerun local DQ2 setup before restarting ganga. Or change inputdata.datasetType to 'local'")
                    raise
            elif job.backend._name=="Local":
                 environment["BACKEND"]="Local"
            else:
                 environment["BACKEND"]="batch"

# now doing output files....

        outfilelist=""
        subjob_outbox=[]
        for type in app.outputpaths.keys():
            outfilelist+=app.outputpaths[type]+app.subjobsOutfiles[job.id][type]+" "
            if job.application.se_name=="ganga":
                outfile1=app.subjobsOutfiles[job.id][type]
                subjob_outbox.append(outfile1)
        environment["OUTPUTFILES"]=outfilelist
        # Work around for glite WMS spaced environement variable problem
        inputbox += [ FileBuffer('outputfiles.conf',environment['OUTPUTFILES']+'\n') ]        


 # setting up job wrapper arguments.       
        args=app.args
        
        jid=""
        if job._getRoot().subjobs:
            jid = job._getRoot().id
        else:
            jid = "%d" % job.id
        environment["OUTPUT_JOBID"]=str(jid) # used for versionning
        if app.dryrun:
            environment["DRYRUN"] = "TRUE"
        if app.dbrelease:
            environment["ATLASDBREL"]=app.dbrelease
        inputdata = []

        filename="wrapper.sh"
        exe = os.path.join(os.path.dirname(__file__),filename)

#       output sandbox
        outputbox =jobmasterconfig.outputbox
        outputbox.extend(subjob_outbox)

        return StandardJobConfig(File(exe),inputbox,args,outputbox,environment) 



allHandlers.add('AthenaMC','Local',AthenaMCLocalRTHandler)
allHandlers.add('AthenaMC','LSF',AthenaMCLocalRTHandler)
#allHandlers.add('AthenaMC','Condor',AthenaMCLocalRTHandler) # condor has requirements...
allHandlers.add('AthenaMC','PBS',AthenaMCLocalRTHandler)
allHandlers.add('AthenaMC','SGE',AthenaMCLocalRTHandler)


config = getConfig('AthenaMC')
logger = getLogger('AthenaMC')
